package org.openxma.dsl.generator.helper;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.openxma.dsl.core.CoreFactory;
import org.openxma.dsl.core.model.ReferenceableByXmadslVariable;
import org.openxma.dsl.core.model.StatusFlag;
import org.openxma.dsl.core.model.XmadslVariable;
import org.openxma.dsl.pom.PomFactory;
import org.openxma.dsl.pom.model.Command;
import org.openxma.dsl.pom.model.Component;
import org.openxma.dsl.pom.model.ComposeData;
import org.openxma.dsl.pom.model.ComposedElement;
import org.openxma.dsl.pom.model.Content;
import org.openxma.dsl.pom.model.CustomizationOfComposite;
import org.openxma.dsl.pom.model.CustomizationOfGuiElement;
import org.openxma.dsl.pom.model.CustomizeableField;
import org.openxma.dsl.pom.model.DataObjectVariable;
import org.openxma.dsl.pom.model.EventFunction;
import org.openxma.dsl.pom.model.EventMapping;
import org.openxma.dsl.pom.model.GuiElementEventMapping;
import org.openxma.dsl.pom.model.IGuiElementWithEvent;
import org.openxma.dsl.pom.model.InitEventMapping;
import org.openxma.dsl.pom.model.Page;
import org.openxma.dsl.pom.model.SetOfGuiElements;
import org.openxma.dsl.pom.model.Table;
import org.openxma.dsl.pom.model.TableColumn;
import org.openxma.dsl.pom.model.XMAWidgetEventMapping;
import org.openxma.dsl.pom.model.XmadslComposite;
import org.openxma.dsl.pom.model.XmadslPage;
import org.openxma.dsl.pom.naming.PomNameProvider;
import org.openxma.dsl.pom.scoping.PomDslScopeProvider;
import org.openxma.dsl.pom.scoping.PomElementCollector;

public class InheritePageExtension {

	public static void addInheritedProperties(XmadslPage xmadslPage, Component comp) {
		Page superPage = xmadslPage.getTemplate();
		if (superPage == null) {
			throw new RuntimeException("To add inherited properties to a page the super-page must not be null");
		}
		inheritDataObjectVariables(xmadslPage, superPage);
		inheritStatusFlags(xmadslPage, superPage);
		inheritCommands(xmadslPage, superPage);
		inheritContent(xmadslPage, superPage, comp);
		inheritEventMapping(xmadslPage, superPage);

		// Update references
		List<EObject> allElementsOfPage = PomDslScopeProvider.getNsElementsInRange(xmadslPage, null);
		updateReferences(xmadslPage, superPage, comp, allElementsOfPage);
	}

	private static void inheritContent(XmadslPage xmadslPage, Page superPage, Component comp) {
		// Copy the content of the super page
		Content copyOfSuperContent = (Content)EcoreUtil.copy(superPage.getContent());
		xmadslPage.setContent(copyOfSuperContent);

		List<EObject> elementsOfPage = PomDslScopeProvider.getNsElementsInRange(xmadslPage.getContent(), null);

		// Customize the content of the super page
		Iterator<CustomizationOfGuiElement> exIt = xmadslPage.getExtensions().iterator();
		while (exIt.hasNext()) {
			CustomizationOfGuiElement cust = exIt.next();
			// Composite
			if (cust instanceof CustomizationOfComposite) {
				CustomizationOfComposite cc = (CustomizationOfComposite)cust;
				if (cc.getComposite() instanceof XmadslComposite) {
					XmadslComposite xc = (XmadslComposite)cc.getComposite();
					String compositeName = xc.getName();
					// Find composite in content
					for (int i=0;i<elementsOfPage.size();i++) {
						EObject obj = elementsOfPage.get(i);
						if (obj instanceof XmadslComposite && ((XmadslComposite)obj).getName().equals(compositeName)) {
							XmadslComposite targetComposite = (XmadslComposite)obj;
							// Apply Customization...
							if (targetComposite.getContent()==null) {
								targetComposite.setContent(PomFactory.eINSTANCE.createContent());
							}
							inheritCompositeContent(targetComposite.getContent(),cc.getAddedWidgetSetItems(),cc.getAddedElements());
							if (cc.getComposeLayout() != null && cc.getComposeLayout().getTabulators().size()>0) {
								targetComposite.getContent().setComposeLayout((ComposeData)EcoreUtil.copy(cc.getComposeLayout()));

							}

						}
					}

				}
				continue;
			}
		}
	}

	private static void inheritCompositeContent(Content cc, EList<SetOfGuiElements> setItems, EList<ComposedElement> composedItems) {
		if (setItems.size()>0) {
			cc.getWidgetSetItems().addAll(EcoreUtil.copyAll(setItems));
		}
		if (composedItems.size()>0) {
			cc.getElements().addAll(EcoreUtil.copyAll(composedItems));
		}
	}

	private static String determineId(EObject obj) {
		return PomNameProvider.getXmaName(obj);
	}

	private static abstract class ReferenceUpdater<TypeOfReferencedObject extends EObject> {
		EObject referencingObj;
		TypeOfReferencedObject referencedObj;
		Page xmadslPage;
		Page superPage;

		ReferenceUpdater(EObject referencingObj, TypeOfReferencedObject referencedObj, Page xmadslPage, Page superPage) {
			this.referencingObj = referencingObj;
			this.referencedObj = referencedObj;
			this.xmadslPage = xmadslPage;
			this.superPage = superPage;
		}
		String getIdOfReferencedObject() {
			return determineId(referencedObj);
		}
		EObject getReferenceObject(Collection<? extends EObject> referenceableObjects) {
			String idOfReferencedObj = getIdOfReferencedObject();
			if (idOfReferencedObj == null) {
				throw new RuntimeException("Id of referenced object is null");
			}
			for (EObject referenceableObj : referenceableObjects) {
				String id = PomNameProvider.getXmaName(referenceableObj);
				if (idOfReferencedObj.equals(id)) {
					return referenceableObj;
				}
			}
			if (referencedObj.equals(superPage)) {
				return xmadslPage;
			}
			return null;
		}
		abstract void update(TypeOfReferencedObject referencedObj);
	}

	private static void updateReference(ReferenceUpdater refUpdater, Collection<? extends EObject> referenceableObjects) {
		EObject refObj = refUpdater.getReferenceObject(referenceableObjects);
		if (refObj != null) {
			refUpdater.update(refObj);
		}
	}

	private static void updateReferences(
			XmadslPage xmadslPage, Page superPage, Component comp, List<EObject> allElementsOfPage) {

		Collection<EObject> allDataObjectVariables = PomDslScopeProvider.getNsDataObjectVariable(comp, xmadslPage);

		for (EObject obj : allElementsOfPage) {

			if (obj instanceof XmadslVariable) {
				XmadslVariable xv = (XmadslVariable)obj;
				updateReference(new ReferenceUpdater<ReferenceableByXmadslVariable>(obj,xv.getReference(),xmadslPage,superPage) {
					void update(ReferenceableByXmadslVariable referencedObj) {
						((XmadslVariable)referencingObj).setReference(referencedObj);
					}
				}, allElementsOfPage);

			}
			// CustomizeableFields
			if (obj instanceof CustomizeableField) {
				updateReference(new ReferenceUpdater<DataObjectVariable>(obj,((CustomizeableField)obj).getObject(),xmadslPage,superPage) {
					void update(DataObjectVariable referencedObj) {
						((CustomizeableField)referencingObj).setObject(referencedObj);
					}
				}, allDataObjectVariables);
				continue;
			}
			// Table
			if (obj instanceof Table) {
				updateReference(new ReferenceUpdater<DataObjectVariable>(obj,((Table)obj).getObject(),xmadslPage,superPage) {
					void update(DataObjectVariable referencedObj) {
						((Table)referencingObj).setObject(referencedObj);
					}
				}, allDataObjectVariables);
				continue;
			}
			// TableColumn
			if (obj instanceof TableColumn) {
				updateReference(new ReferenceUpdater<DataObjectVariable>(obj,((TableColumn)obj).getObject(),xmadslPage,superPage) {
					void update(DataObjectVariable referencedObj) {
						((TableColumn)referencingObj).setObject(referencedObj);
					}
				}, allDataObjectVariables);
				continue;
			}
			// EventMapping - EventFunction (Command)
			if (obj instanceof EventMapping) {

				//List<Command> commands = org.eclipse.xtext.EcoreUtil2.typeSelect(allElementsOfPage, Command.class);

				Collection<Command> commands = PomElementCollector.getAccessibleCommands(xmadslPage,comp);

				//((EventMapping)obj).getEventFunctions().clear();	// Todo check
				for (EventFunction eventFunction : ((EventMapping)obj).getEventFunctions()) {
					updateReference(new ReferenceUpdater<Command>(eventFunction,eventFunction.getCommand(),xmadslPage,superPage) {
						void update(Command referencedObj) {
							((EventFunction)referencingObj).setCommand(referencedObj);
						}
					}, commands);
				}


				// GuiElementEventMapping - Control
				if (obj instanceof GuiElementEventMapping) {
					updateReference(new ReferenceUpdater<IGuiElementWithEvent>(obj,((GuiElementEventMapping)obj).getControl(),xmadslPage,superPage) {
						void update(IGuiElementWithEvent referencedObj) {
							((GuiElementEventMapping)referencingObj).setControl(referencedObj);
						}
					}, commands);
				}
				continue;
			}
		}
	}

	private static void inheritDataObjectVariables(XmadslPage xmadslPage, Page superPage) {
		// Copy commands
		Collection<DataObjectVariable> dos = (Collection<DataObjectVariable>)EcoreUtil.copyAll(superPage.getDataobjects());
		xmadslPage.getDataobjects().addAll(dos);
	}

	private static void inheritCommands(XmadslPage xmadslPage, Page superPage) {
		// Copy commands
		Collection<Command> commands = (Collection<Command>)EcoreUtil.copyAll(superPage.getCommands());
		xmadslPage.getCommands().addAll(commands);
	}

	private static void inheritStatusFlags(XmadslPage xmadslPage, Page superPage) {

		if (superPage.getConditionsBlock() == null) {
			return;
		}

		// Copy statusFlags
		Collection<StatusFlag> statusFlags = (Collection<StatusFlag>)EcoreUtil.copyAll(superPage.getConditionsBlock().getStatusFlags());
		if (xmadslPage.getConditionsBlock() == null) {
			xmadslPage.setConditionsBlock(CoreFactory.eINSTANCE.createConditionsBlock());
		}
		xmadslPage.getConditionsBlock().getStatusFlags().addAll(statusFlags);
	}

	private static void inheritEventMapping(XmadslPage xmadslPage, Page superPage) {

		// Copy eventmapping
		Collection<EventMapping> superEvents = (superPage.getEvents() != null) ? (Collection<EventMapping>)EcoreUtil.copyAll(superPage.getEvents().getMapping()):null;
		Collection<EventMapping> ownEvents = (xmadslPage.getEvents() != null) ? (Collection<EventMapping>)EcoreUtil.copyAll(xmadslPage.getEvents().getMapping()):null;

		if (superEvents != null) {
			for (EventMapping superMapping : superEvents) {
				if (superMapping instanceof InitEventMapping) {
					// Override command
					if (ownEvents != null) {
						for (EventMapping ownMapping : ownEvents) {
							if (ownMapping instanceof InitEventMapping) {
								superMapping.getEventFunctions().addAll(ownMapping.getEventFunctions());
								ownEvents.remove(ownMapping); // Remove mappings which exists already in superPage
								break;
							}
						}
					}
				} else if (superMapping instanceof GuiElementEventMapping) {
					GuiElementEventMapping superGuiElementEventMapping = (GuiElementEventMapping)superMapping;
					String idOfSuperControl = PomNameProvider.getXmaName(superGuiElementEventMapping.getControl());
					// Override command
					if (ownEvents != null) {
						for (EventMapping ownMapping : ownEvents) {
							if (ownMapping instanceof GuiElementEventMapping) {
								GuiElementEventMapping ownGuiElementEventMapping = (GuiElementEventMapping)ownMapping;
								String idOfOwnControl = PomNameProvider.getXmaName(ownGuiElementEventMapping.getControl());
								if (superGuiElementEventMapping.getEvent().equals(ownGuiElementEventMapping.getEvent())
										&& idOfSuperControl.equals(idOfOwnControl)) {
									superGuiElementEventMapping.getEventFunctions().addAll(ownMapping.getEventFunctions());
									ownEvents.remove(ownMapping); // Remove mappings which exists already in superPage
									break;
								}
							}
						}
					}
				} else if (superMapping instanceof XMAWidgetEventMapping) {
					// TODO implement support for XMAWidgetEventMapping
					throw new RuntimeException("Inheritance of XMAWidgetEventMapping is not supported");
				}
			}

			// Add all events which does not already exist in superPage
			if (ownEvents != null && ownEvents.size()>0) {
				superEvents.addAll(ownEvents);
			}

			// Set the inherited EventMapping
			if (xmadslPage.getEvents()==null) {
				xmadslPage.setEvents(PomFactory.eINSTANCE.createEventMappingList());
			}
			xmadslPage.getEvents().getMapping().clear();
			xmadslPage.getEvents().getMapping().addAll(superEvents);
		}

	}

}

