/*
 * @(#) $Id: ServiceExtension.java 9761 2012-07-13 13:22:22Z mjohannes $
 *
 * Copyright 2009/2010 by sIT Solutions,
 * A-1110 Wien, Geiselbergstr.21-25.
 * All rights reserved.
 *
 */
package org.openxma.dsl.generator.helper;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.xtext.EcoreUtil2;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.core.model.ModelElement;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.CrudOperationType;
import org.openxma.dsl.dom.model.Dao;
import org.openxma.dsl.dom.model.DataView;
import org.openxma.dsl.dom.model.DelegateOperation;
import org.openxma.dsl.dom.model.Operation;
import org.openxma.dsl.dom.model.Service;

import com.google.common.collect.Lists;

public class ServiceExtension {

    static class CreateUpdateOperations {
        Operation create;
        Operation update;
    }

    public static List<Operation> getCreateUpdateOperations(List<Operation> in) {
        Map<String,CreateUpdateOperations> opTypeToCUFOpsMap = new HashMap<String,CreateUpdateOperations>();
        for (Operation operation : in) {
            DelegateOperation delegateOp = operation.getDelegate();
            if ( delegateOp != null ) {
                CrudOperationType t = delegateOp.getCrudOperationType();
                if ( t != null && t == CrudOperationType.CREATE || t == CrudOperationType.UPDATE ) {
                    String typeName = delegateOp.getViewParameter().getName();
                    CreateUpdateOperations ops4type = opTypeToCUFOpsMap.get(typeName);
                    if ( ops4type == null ) {
                        ops4type = new CreateUpdateOperations();
                        opTypeToCUFOpsMap.put(typeName,ops4type);
                    }
                    if ( t == CrudOperationType.CREATE )
                        ops4type.create = operation;
                    else if ( t == CrudOperationType.UPDATE )
                        ops4type.update = operation;
                }
            }
        }
        List<Operation> cuOperations = Lists.newArrayList();
        for ( CreateUpdateOperations ops4type : opTypeToCUFOpsMap.values() ) {
            if ( ops4type.create != null && ops4type.update != null )
                cuOperations.add(ops4type.create);
        }
        return cuOperations;
    }

    public static Attribute getDataViewVersion(DataView view) {
        if ( view == null )
            return null;
        Attribute res = getDataViewVersion(view.getSuperType());
        if ( res != null )
            return res;
        for ( Attribute attr : view.getAllAttributes() )
            if ( attr.isVersion() )
                return attr;
        return null;
    }

    public static Attribute getDataViewId(DataView view) {
        if ( view == null )
            return null;
        Attribute res = getDataViewId(view.getSuperType());
        if ( res != null )
            return res;
        for ( Attribute attr : view.getAllAttributes() )
            if ( attr.isIdentifier() )
                return attr;
        return null;
    }

    public static Attribute getDataViewVersionOrId(DataView view) {
        Attribute res = getDataViewVersion(view);
        if ( res == null )
            res = getDataViewId(view);
        return res;
    }

    public static Set<Service> getAllServicesFromResourceSet(Model model) {
        return getAllElemeentsFromResourceSet(model, Service.class);
    }
    public static Set<Dao> getAllDaosFromResourceSet(Model model) {
        return getAllElemeentsFromResourceSet(model, Dao.class);
    }
    public static Set<DataView> getAllDataViewsFromResourceSet(Model model) {
        return getAllElemeentsFromResourceSet(model, DataView.class);
    }

    public static <T extends ModelElement> Set<T> getAllElemeentsFromResourceSet(Model model, Class<T> type) {
        Set<T> modelElements = new HashSet<T>();
        Iterator<Resource> iterator = model.eResource().getResourceSet().getResources().iterator();
        while (iterator.hasNext()) {
            Resource resource = (Resource) iterator.next();
            EObject eObject = resource.getEObject("/");
            if (eObject instanceof Model) {
                List<? extends ModelElement> ofType = EcoreUtil2.getAllContentsOfType(eObject,type);
                modelElements.addAll((List<? extends T>) ofType);
            }
        }
        return modelElements;
    }
}
