package org.openxma.dsl.generator.helper;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.openxma.dsl.common.Strings;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.core.model.Type;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.ComplexType;
import org.openxma.dsl.dom.model.DataView;
import org.openxma.dsl.dom.model.PresentableFeature;
import org.openxma.dsl.dom.model.ValueObject;
import org.openxma.dsl.pom.model.ComposedElement;
import org.openxma.dsl.pom.model.Composite;
import org.openxma.dsl.pom.model.DataObjectVariable;
import org.openxma.dsl.pom.model.FieldFeature;
import org.openxma.dsl.pom.model.Page;
import org.openxma.dsl.pom.model.ReferencedXMAComposite;
import org.openxma.dsl.pom.scoping.PomElementCollector;

import at.spardat.xma.guidesign.BDAttribute;
import at.spardat.xma.guidesign.BusinessData;
import at.spardat.xma.guidesign.GuidesignFactory;
import at.spardat.xma.guidesign.IBDAttachable;
import at.spardat.xma.guidesign.IEditable;
import at.spardat.xma.guidesign.IFormaterAttachable;
import at.spardat.xma.guidesign.IWidgetWithLabel;
import at.spardat.xma.guidesign.State;
import at.spardat.xma.guidesign.ValidInState;
import at.spardat.xma.guidesign.XMACombo;
import at.spardat.xma.guidesign.XMAComponent;
import at.spardat.xma.guidesign.XMAComposite;
import at.spardat.xma.guidesign.XMADialogPage;
import at.spardat.xma.guidesign.XMALabel;
import at.spardat.xma.guidesign.XMAPage;
import at.spardat.xma.guidesign.XMAWidget;
import at.spardat.xma.guidesign.flex.CompoundValidator;
import at.spardat.xma.guidesign.flex.FlexFactory;
import at.spardat.xma.guidesign.flex.ICanBeMandatory;
import at.spardat.xma.guidesign.flex.IDefSelectFunctionCaller;
import at.spardat.xma.guidesign.flex.ISelectFunctionCaller;
import at.spardat.xma.guidesign.flex.MenuItem;
import at.spardat.xma.guidesign.flex.XMAFunction;
import at.spardat.xma.guidesign.types.BeanBaseType;

/**
 * Utility for extending a field
 * TODO rename to XmaGeneratorUtil 
 */
public class Util {

	// private static final Log logger = LogFactory.getLog(Util.class);
	
	/**
	 * Get first XMAWidget of xmaPage where namInstance equals name
	 */
	public static XMAWidget getWidgetForName(XMAPage xmaPage, String name) {		
		TreeIterator<EObject> allContentIterator = xmaPage.eAllContents();
		while (allContentIterator.hasNext()) {
			EObject obj = allContentIterator.next();
			if (obj instanceof XMAWidget) {
				XMAWidget w = (XMAWidget)obj;
				String namInstance = w.getNamInstance();
				if (namInstance != null && namInstance.equals(name)) {
					return w;
				}
			}
		}
		return null;		
	}

    public static MenuItem getMenuItemForName(XMAPage xmaPage, String name) {
        TreeIterator<EObject> allContentIterator = xmaPage.eAllContents();
        while (allContentIterator.hasNext()) {
            EObject obj = allContentIterator.next();
            if (obj instanceof MenuItem) {
                MenuItem m = (MenuItem)obj;
                String menuName = m.getNamRscKeyLabel();
                if (menuName != null && menuName.equals(name)) {
                    return m;
                }
            }
        }
        return null;
    }
    
	
	/**
	 *
	 * @param label
	 * @param text
	 * @return
	 */
	public static XMALabel setLabelText(XMALabel label, String text) {
		if (text != null && text.length()>0) {
			label.setRscLabel(text);
		}
		return label;
	}

	/**
	 *
	 * @param widget
	 * @param text
	 * @return
	 */
	public static XMAWidget setTooltipText(XMAWidget widget, String text) {
		if (text != null && text.length()>0) {
			widget.setRscTooltip(text);
		}
		return widget;
	}

	/**
	 *
	 * @param widget
	 * @param label
	 * @return
	 */
	public static XMAWidget attachWidgetAndLabel(XMAWidget widget,XMALabel label) {

		if (widget instanceof IWidgetWithLabel) {
			label.setLabelWidget((IWidgetWithLabel)widget);
			((IWidgetWithLabel)widget).setLabel(label);
		} else {
			throw new RuntimeException("Widget does not support label");
		}

		return widget;
	}

	/**
	 * Attach a data attribute to a widget.
	 *
	 * @param widget
	 * @param bdAttribute
	 * @return
	 */
	public static XMAWidget setBDAttribute(XMAWidget widget, BDAttribute bdAttribute) {

		if (widget instanceof IBDAttachable) {
			((IBDAttachable)widget).setDataAttribute(bdAttribute);
		} else {
			throw new RuntimeException("Widget does not support attaching businessdata attributes");
		}

		return widget;
	}

	/**
	 *
	 * @param xmaPage
	 * @return
	 * TODO: check and refactore duplicate implementation: getContainingXMAPage in Guidesign.ext
	 */
	public static XMAPage getSuperPage(XMAPage xmaPage) {

		if (xmaPage != null) {
			EObject eObject = xmaPage.eContainer();
			while (eObject != null) {
				if (eObject instanceof XMAPage) {
                    return (XMAPage)eObject;
                }
				eObject = eObject.eContainer();
			}
		}
		return null;
	}

	/**
	 *  Get a list of all BusinessData which are can be accessed from the XMAPage
	 * @param page
	 * @return
	 */
	public static List<BusinessData> getAccessibleBusinessData(XMAPage page) {

		List<BusinessData> list = new ArrayList<BusinessData>();
		list.addAll(page.getBusinessDataVariables()); // Copy
		XMAPage superPage = getSuperPage(page);

		if (superPage != null) {
			list.addAll(getAccessibleBusinessData(superPage));
		} else {
			list.addAll(page.getComponent().getBusinessDataVariables());
		}
		return list;
	}	
	
	public static BDAttribute findBDAttributeForAttribute(XMAPage xmaPage, PresentableFeature f, DataObjectVariable dataObject) {		
		List accessibleBusinessData = getAccessibleBusinessData(xmaPage);
		return findBDAttributeForAttribute(xmaPage.getComponent(),accessibleBusinessData,f,dataObject);		
	}
	
	public static BDAttribute findBDAttributeForAttribute(XMAComponent xmaComp, List<BusinessData> businessDataList, PresentableFeature f, DataObjectVariable dataObject) {		
		BusinessData bd = findOrCreateBusinessData(xmaComp,businessDataList,dataObject.getType(),dataObject.getName(),dataObject.isIsCollection());		
		if (bd == null) {
			return null;
		}
		Attribute attr = f.getAttribute();
		Type type = attr.getType().getDataType();
		if (type instanceof SimpleType) {
			return findAttributeOfBusinessData(bd,f);
		} 		
		return null;
	}	
	
	/**
	 *
	 * @param xmaPage
	 * @param attribute
	 * @param dataObject
	 * @return
	 */
	public static BDAttribute findBDAttributeForFeature(XMAPage xmaPage, FieldFeature fieldFeature, DataObjectVariable dataObject) {
		List accessibleBusinessData = getAccessibleBusinessData(xmaPage);
		return findBDAttributeForFeature(xmaPage.getComponent(),accessibleBusinessData,fieldFeature,dataObject);
	}

	/**
	 *
	 * @param businessDataList
	 * @param attribute
	 * @param dataObject
	 * @return
	 */
	public static BDAttribute findBDAttributeForFeature(XMAComponent xmaComp, List<BusinessData> businessDataList, FieldFeature fieldFeature, DataObjectVariable dataObject) {		
		BusinessData rootBusinessData = findOrCreateBusinessData(xmaComp,businessDataList,dataObject.getType(),dataObject.getName(),dataObject.isIsCollection());		
		if (rootBusinessData == null) {
			return null;
		}
		return findBDAttributeForFieldFeature(xmaComp,businessDataList,rootBusinessData,fieldFeature,dataObject.isIsCollection());
	}
	
	
	private static BDAttribute findBDAttributeForFieldFeature(XMAComponent xmaComp, List<BusinessData> businessDataList,BusinessData bd,FieldFeature fieldFeature, boolean isCollection) {
		
		PresentableFeature f = fieldFeature.getFeature();
		String bdName = bd.getNamInstance();
		// TODO: use this bdName as prefix for the name of e.g. the name of a ValueObject in order to allow
		// to distinguish "sub"-businessdatas from each other. But in this case the name of the getter in the
		// parent object is different from the new "full"-name and this is currently not supported from openXMA core.
		// -> Redesign the naming and dataAccess in openXMA core. 
		if (f.getAttribute()!=null && !f.getAttribute().isReference()) {
			Attribute a = f.getAttribute();
			Type type = a.getType().getDataType();
			if (type instanceof SimpleType) {
				return findAttributeOfBusinessData(bd,f);
			} else if (type instanceof ValueObject) {
				bdName=a.getName();	
				boolean isAttributeCollection = a.isMany();
				if (isAttributeCollection) throw new RuntimeException();
				BusinessData fd = findOrCreateBusinessData(xmaComp,businessDataList,(ValueObject)type,bdName,isAttributeCollection);
				fd.setIsAttributeOf(bd);
				bd=fd;
			} else if (type instanceof DataView) {
				bdName=a.getName();	
				boolean isAttributeCollection = a.isMany();				
				BusinessData fd = findOrCreateBusinessData(xmaComp,businessDataList,(DataView)type,bdName,isAttributeCollection);
				fd.setIsAttributeOf(bd);
				bd=fd;				
			} else {
				throw new RuntimeException("Unknown type of AttributeHolder: "+f);
			}
		}
		if (f.getAttribute()!=null && f.getAttribute().isReference()) {
			bdName=f.getName();				
			boolean isAttributeCollection = f.getAttribute().isMany();
			BusinessData fd = findOrCreateBusinessData(xmaComp,businessDataList,(ComplexType) f.getAttribute().getDataType(),bdName,isAttributeCollection);
			fd.setIsAttributeOf(bd);
			bd=fd;								
		}
		if (fieldFeature.getSubFeature() != null) {
			return findBDAttributeForFieldFeature(xmaComp,businessDataList, bd,fieldFeature.getSubFeature(),false); // TODO implement support for collections
		}		
		return null;
	}
	
	private static BDAttribute findAttributeOfBusinessData(BusinessData bd, PresentableFeature f) {
		List<BDAttribute> bdAttributes = bd.getAttributes();
		for (BDAttribute bdAttr: bdAttributes) {
			if (bdAttr.getNamAttrVal().equals(f.getCalculatedName())) {
				return bdAttr;
			}
		}
		return null;
	}
	
	public static BusinessData findExistingBusinessData(List<BusinessData> businessDataList, ComplexType type, String name, boolean isCollection) {
		for (BusinessData bd : businessDataList) {									
			if (bd.getNamInstance().equals(name)) {
				return bd;
			}
		}
		return null;
	}	
	
	public static BusinessData findOrCreateBusinessData(XMAComponent xmaComp, List<BusinessData> businessDataList, ComplexType type, String name, boolean isCollection) {
		// Try to find an existing one
		BusinessData foundBd = findExistingBusinessData(businessDataList,type,name,isCollection);
		if (foundBd != null) {
			return foundBd;
		}
		// Not found - create a new one
		BusinessData bd = createBusinessData(type,xmaComp,name,isCollection);
		if (bd != null) {
			businessDataList.add(bd);
		} else {
			throw new RuntimeException("XMA-BusinessData not found or created: '"+name+"'");			
		}
		return bd;
	}	
	
	// Create BusinessData and add it to businessdataCol of xmaComp  	
	public static BusinessData createBusinessData(ComplexType type, XMAComponent xmaComp, String name, Boolean isCollection) {

		if (xmaComp == null) {
			throw new IllegalArgumentException("The parameter xmaComp must not be null");
		}
		String fullClassName = NameExtension.getFullClassNameForDatamapping(type);
		
		// TODO: derive BeanBaseType from baseClass 
		BeanBaseType beanType = BeanBaseType.OTHER_LITERAL;
		
		if (xmaComp.getBusinessdataCol() == null) {
			xmaComp.setBusinessdataCol(GuidesignFactory.eINSTANCE.createBDCollection());
		}
			
		BusinessData bd = GuidesignFactory.eINSTANCE.createBusinessData();		
		bd.setNamBDClass(fullClassName);
		bd.setNamBaseType(beanType);
		bd.setNamInstance(name);
		bd.setYnGenerated(true);
		bd.setYnCollection(isCollection);
		for (PresentableFeature attribute : type.getAllPresentableFeaturesIncludingSuperType()) {
			BDAttribute bda = createBDAttribute(attribute,bd);
			if (bda != null) {
				bd.getAttributes().add(bda);
			}
		}		
		// Add BusinessData to XMACompnent
		xmaComp.getBusinessdataCol().getBusinessdata().add(bd);						
		return bd;	
	}
	
	public static BDAttribute createBDAttribute(PresentableFeature pf, BusinessData bd) {
		
		BDAttribute bda = GuidesignFactory.eINSTANCE.createBDAttribute();
		
		Attribute attr = pf.getAttribute();
	
		// temp fix to work with BDAttributeImpl (expects fully qualified type names even for java.lang packages)  
		String propertyType = NameExtension.getFullyQualifiedName(attr.getType(),true);
		String name =  pf.getCalculatedName();
		bda.setNamAttrVal(name);
		bda.setNamPropertyType(propertyType.contains(".") ?  propertyType : "java.lang."+propertyType);	
		String nameFirstUpper = Strings.capitalize(name);
		if (attr.isReadOnly() == false) {		
			bda.setNamSetter("set"+nameFirstUpper);
		}
		bda.setNamGetter("get"+nameFirstUpper);	
		//this.setNamSetterParamType(attr.getJavaType())->	

		return bda;
	}



	/**
	 * Calculates the digits before comma and returns it as String
	 *
	 * @param precision
	 * @param scale
	 * @return
	 */
	public static String calculateBeforeComma(String precision, String scale) {

		int p = Integer.parseInt(precision);
		int s = Integer.parseInt(scale);

		int beforeComma = p-s;

		if (beforeComma<0) {
			throw new RuntimeException("Precision of a number must be equal or greater then scale");
		}

		return Integer.toString(beforeComma);
	}

	/**
	 * Try to cast the XMAWidget to ISelectFunctionCaller and set the
	 * selectFunction.
	 *
	 * @param w
	 * @param f
	 * @return
	 */
	public static XMAWidget castAndSetSelectFunction(XMAWidget w, XMAFunction f) {
		if (w instanceof ISelectFunctionCaller) {
			((ISelectFunctionCaller)w).getSelectFunction().add(f);
		} else {
			throw new RuntimeException("Type of widget does not support 'selectFunction'");
		}
		return w;
	}

	/**
	 * Try to cast the XMAWidget to IDefSelectFunctionCaller and set the
	 * defaultSelectFunction.
	 *
	 * @param w
	 * @param f
	 * @return
	 */
	public static XMAWidget castAndSetDefSelectFunction(XMAWidget w, XMAFunction f) {
		if (w instanceof IDefSelectFunctionCaller) {
			((IDefSelectFunctionCaller)w).getDefSelectFunction().add(f);
		} else {
			throw new RuntimeException("Type of widget does not support 'defSelectFunction'");
		}
		return w;
	}

	/**
	 * Get the initState of the containing DialogPage for a EObject. (Rekursiv)
	 *
	 * @param obj
	 * @return
	 */
	private static State getInitState(EObject obj) {

		State state = null;

		if (obj instanceof XMADialogPage) {
			// DialogPage found
			XMADialogPage dialogPage = (XMADialogPage)obj;

			EList states = dialogPage.getState();
			if (states != null && states.size()>0) {
				state = (State)states.get(0);
			}
		} else {
			// Rekursiv for the next container
			EObject container = obj.eContainer();
			if (container != null) {
				state = getInitState(container);
			}
		}

		return state;
	}

	/**
	 *
	 * @param w
	 * @return
	 */
	public static Boolean isFormatterAttachable(EObject w) {

		if (w instanceof IFormaterAttachable) {
			return true;
		}
		return false;
	}

	/**
	 * Cast to IFormaterAttachable
	 * @param widget
	 * @return
	 */
	public static IFormaterAttachable castToIFormaterAttachable(XMAWidget widget) {

		if (widget instanceof IFormaterAttachable) {
			return (IFormaterAttachable)widget;
		}
		return null;
	}
	
	/**
	 * Cast to ICanBeMandatory
	 * @param widget
	 * @return
	 */
	public static ICanBeMandatory castToICanBeMandatory(XMAWidget widget) {

		if (widget instanceof ICanBeMandatory) {
			return (ICanBeMandatory)widget;
		}
		return null;
	}	

	/**
	 * Cast to IEditable
	 * @param widget
	 * @return
	 */
	public static IEditable castToIEditable(XMAWidget widget) {

		if (widget instanceof IEditable) {
			return (IEditable)widget;
		}
		return null;
	}

	/**
	 * Cast to XMAWidget
	 * @param obj
	 * @return
	 */
	public static XMAWidget castToXMAWidget(EObject obj) {

		if (obj instanceof XMAWidget) {
			return (XMAWidget)obj;
		}
		return null;
	}

	/**
	 *
	 * @return
	 */
	public static CompoundValidator attachCompoundValidatorToWidget(EObject w,
			ValidInState validInState) {

		CompoundValidator cv = null;

		if (isFormatterAttachable(w)) {

			cv = FlexFactory.eINSTANCE.createCompoundValidator();

			IFormaterAttachable fma = (IFormaterAttachable)w;
			State initState = getInitState(w);

			validInState.setValidator(cv);
			validInState.setParent(fma);
			validInState.setState(initState);

			fma.getValidator().clear(); // Clear if already exists!
			fma.getValidator().add(validInState);

		} else {
			throw new RuntimeException(
					"Type of widget '"+w+"' does not support attaching a validator");
		}
		return cv;
	}

	/**
	 *
	 * @param combo
	 * @param dataSource
	 * @return
	 */
	public static XMACombo setDataSource(XMACombo combo, String dataSource) {

		combo.setUriDataSource(dataSource);
		return combo;
	}

	public static boolean removeObject(org.eclipse.emf.ecore.EObject obj) {

		//org.eclipse.emf.ecore.util.EcoreUtil.remove(obj);
		org.eclipse.emf.ecore.util.EcoreUtil.delete(obj);
		return true;
	}
	
	/**
	 * Determine the content of a Page is referencing the PageComposite of the XMAPage.
	 * 
	 * @param page
	 * @return
	 */
	public static boolean contentReferencesPageComposite(Page page) {
		if (page.getContent()==null) return false;
		EList<ComposedElement> items = page.getContent().getElements();
		if (items.size()==1 && items.get(0) instanceof ReferencedXMAComposite) {
			ReferencedXMAComposite ref = (ReferencedXMAComposite)items.get(0);
			if (ref.isReferencingPageComposite() == true) {
				return true;
			}
		}
		// GuiElements direct placed into Page without "content" or ReferencedXMAComposite		
		if (items.size()>0 && items.get(0) instanceof ReferencedXMAComposite == false) {
			return true;
		}
		return false;
	}
	
	public static Composite getPageComposite(Page page) {
		EList<ComposedElement> items = page.getContent().getElements();
		if (items.size()==1 && items.get(0) instanceof ReferencedXMAComposite) {
			ReferencedXMAComposite ref = (ReferencedXMAComposite)items.get(0);
			if (ref.isReferencingPageComposite() == true) {
				return ref;
			}
		}
		// GuiElements direct placed into Page without "content" or ReferencedXMAComposite
		if (items.size()>0 && items.get(0) instanceof ReferencedXMAComposite == false) {	
			return page;
		}
		return null;		
	}
	
	public static void addControl(XMAComposite comp, XMAWidget widget, Boolean reverse) {		
		if (reverse != null && reverse) {
			comp.getControls().add(0,widget);
		} else {
			comp.getControls().add(widget);
		}		
	}
	
	public static XMAFunction getXMAFunctionForName(XMAComponent xmaComponent, XMAPage xmaPage, String functionName) {
		Collection<XMAFunction> functionsInScope = PomElementCollector.getInstance().getAccessibleXMAFunctions(xmaPage,xmaComponent);
		for (XMAFunction func : functionsInScope) {
			if (func.getNamFunction().equals(functionName)) {
				return func;
			}			
		}
		return null;				
	}

}

