package org.openxma.dsl.generator.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.openxma.dsl.core.model.DataTypeAndTypeParameter;
import org.openxma.dsl.core.model.Editor;
import org.openxma.dsl.core.model.IntegerParameterValue;
import org.openxma.dsl.core.model.Style;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.generator.LayoutStrategy;
import org.openxma.dsl.generator.helper.FormDataExtension;
import org.openxma.dsl.generator.helper.SimpleTypeExtension;
import org.openxma.dsl.pom.PomFactory;
import org.openxma.dsl.pom.model.Button;
import org.openxma.dsl.pom.model.Composite;
import org.openxma.dsl.pom.model.CustomizeableField;
import org.openxma.dsl.pom.model.FieldPart;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.StylePropertyCharacterWidth;
import org.openxma.dsl.pom.model.StyleSpecification;
import org.openxma.dsl.pom.model.StyleSpecificationProperty;
import org.openxma.dsl.pom.model.TabulatorPosition;
import org.openxma.dsl.pom.util.LayoutData;

import at.spardat.xma.guidesign.GuidesignFactory;
import at.spardat.xma.guidesign.XMAButton;
import at.spardat.xma.guidesign.XMAFormAttachment;
import at.spardat.xma.guidesign.XMAWidget;
import at.spardat.xma.guidesign.types.AttachSideType;


public class DefaultLayoutStrategy implements LayoutStrategy {

    public final static LayoutStrategy instance = new DefaultLayoutStrategy();

    public String getDefaultStyle(EObject obj, String name) {
        return null;
    }

    public int getDefaultGuiElementWidth(GuiElement guiElement, Style style) {

        // Default width for fields
        if (guiElement instanceof CustomizeableField) {
            LayoutData layoutData = new LayoutData((CustomizeableField) guiElement);
            if (layoutData.getRightAttachment() == null) {
                CustomizeableField field = (CustomizeableField)guiElement;
                Attribute attribute = field.getAttribute();
                return getDefaultWidthForType(attribute.getType(),style);
            }
        }
        // Default width for button
        if (guiElement instanceof Button && style != null && "actionButton".equals(style.getName())) {
            return 80;
        }

        return -1;
    }

    public String getEditorImageUri(Editor editor) {
        return null;
    }

    public void setXMAWidgetProperties(XMAWidget widget, GuiElement guiElement, String style) {
    }

    private StylePropertyCharacterWidth getStylePropertyDefaultWidth(Style style) {
        if (style instanceof StyleSpecification) {
            StyleSpecification styleSpec = (StyleSpecification)style;
            for (StyleSpecificationProperty styleProp : styleSpec.getStyleProperty()) {
                if (styleProp instanceof StylePropertyCharacterWidth) {
                    return (StylePropertyCharacterWidth)styleProp;
                }
            }
        }
        return null;
    }

    protected int getDefaultWidthForType(DataTypeAndTypeParameter type, Style style) {

        // Date
        DataTypeAndTypeParameter dateType = SimpleTypeExtension.getTypeForName(type,"Date");
        if (dateType != null) {
            return 90;
        }

        // String
        DataTypeAndTypeParameter stringType = SimpleTypeExtension.getTypeForName(type,"String");
        if (stringType != null) {
            EList parameters = stringType.getParameterValues();
            if (parameters != null && parameters.size()==1 && parameters.get(0) instanceof IntegerParameterValue) {
                IntegerParameterValue maxSizeParameter = (IntegerParameterValue)parameters.get(0);
                StylePropertyCharacterWidth widthStyle = getStylePropertyDefaultWidth(style);
                if (widthStyle != null) {
                    if (widthStyle.getCharWidth()>0) {
                        return Math.max(0,maxSizeParameter.getIntValue())*widthStyle.getCharWidth(); // negative values are set to 0
                    }
                } else if (style == null) {
                    return Math.max(0,maxSizeParameter.getIntValue())*7; // negative values are set to 0
                }
                return -1;
            }
            return 0;
        }
        return 0;
    }

    public List<TabulatorPosition> getTabulatorPositions(Composite composite, List<TabulatorPosition> definedTabulatorPositions, String style) {

        List<TabulatorPosition> tabs = null;

        // Default
        if (style != null && "dataContainer".equals(style)) {
            tabs = new ArrayList<TabulatorPosition>();
            // Left of main control
            TabulatorPosition tab1 = PomFactory.eINSTANCE.createTabulatorPosition();
            tab1.setWidth(130); // 130 pixel
            tabs.add(tab1);
             // Right of main control
//            TabulatorPosition tab2 = PomFactory.eINSTANCE.createTabulatorPosition();
//            tab2.setRelative(true);
//            tab2.setWidth(50); // 50% (% because of relative=true)
//            Offset offset = PomFactory.eINSTANCE.createOffset();
//            offset.setValue(10); // 10 pixel offset
//            tab2.setOffset(offset);
//            tabs.add(tab2);
        }
        return tabs;
    }

    public Boolean useLeftTabInDefaultAttachment(XMAWidget widget, FieldPart fieldPart, String style) {
        if (FieldPart.CONTROL.equals(fieldPart) && FormDataExtension.isButton(widget)==false) return true;
        if (style != null) {
            if ("variableFieldWidth".equals(style)) {
                if (FieldPart.LABEL.equals(fieldPart) && FormDataExtension.isButton(widget)==false) return true;
            }
            if ("useLeftTab".equals(style)) {
                return true;
            }
            return null;
        }
        return false;
    }

    public Boolean useRightTabInDefaultAttachment(XMAWidget widget, FieldPart fieldPart, String style) {
        if (FieldPart.LABEL.equals(fieldPart)) return true;
        if (style != null) {
            if ("variableFieldWidth".equals(style)) {
                if (FieldPart.CONTROL.equals(fieldPart)) return false;
            }
            if ("useRightTab".equals(style)) {
                return true;
            }
            return null;
        }
        if (widget instanceof XMAButton) return false;
        if (FieldPart.CONTROL.equals(fieldPart)) return true;
        return false;
    }

    public XMAFormAttachment getDefaultLeftAttachmentInSet(XMAWidget widget,
            TabulatorPosition tab, XMAWidget prevWidget, XMAWidget nextWidget,
            boolean isLabel, boolean isMainWidget) {
        // TODO Auto-generated method stub
        return null;
    }

    public XMAFormAttachment getDefaultRightAttachmentInSet(XMAWidget widget,
            TabulatorPosition tab, XMAWidget prevWidget, XMAWidget nextWidget,
            boolean isLabel, boolean isMainWidget) {
        // TODO Auto-generated method stub
        return null;
    }

    public void setCustomXmadslModelElementProperties(EObject obj, String name) {
    }

    public boolean isCheckBoxLeftAligned() {
        return true;
    }

    public boolean isVerticalCenteredToPrevious() {
        return false;
    }

    public XMAFormAttachment getDefaultTopAttachmentInSet(XMAWidget xmaWidget,
            XMAWidget prevWidget, XMAWidget centeredWidget,
            XMAWidget widgetForControlFieldPart, Collection<Style> styleList) {

        if (prevWidget != null && isVerticalCenteredToPrevious()) {
            XMAFormAttachment formAttachment = GuidesignFactory.eINSTANCE.createXMAFormAttachment();
            formAttachment.setCodAttachSide(AttachSideType.CENTER_LITERAL);
            formAttachment.setAttachElement(prevWidget);
            return formAttachment;
        }

        return null;
    }

}
