package org.openxma.dsl.generator.impl;

import static org.openxma.dsl.generator.helper.NameExtension.getFullyQualifiedName;
import static org.openxma.dsl.generator.helper.NameExtension.getName;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EObjectContainmentEList;
import org.openxma.dsl.common.Strings;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.dom.model.Dependant;
import org.openxma.dsl.generator.helper.BasicTypeHelper;
import org.openxma.dsl.generator.helper.EventExtension;
import org.openxma.dsl.generator.helper.WidgetExtension;
import org.openxma.dsl.generator.impl.substrategy.BrowseModeStrategy;
import org.openxma.dsl.generator.impl.substrategy.EffectStrategy;
import org.openxma.dsl.generator.impl.substrategy.InvokeStrategy;
import org.openxma.dsl.generator.impl.substrategy.TableColumnVisibilityStrategy;
import org.openxma.dsl.generator.impl.substrategy.TableCustomizerStrategy;
import org.openxma.dsl.generator.impl.substrategy.TableLogicStrategy;
import org.openxma.dsl.pom.model.CallStatement;
import org.openxma.dsl.pom.model.Command;
import org.openxma.dsl.pom.model.CommandImplementation;
import org.openxma.dsl.pom.model.CustomizeableField;
import org.openxma.dsl.pom.model.EventFunction;
import org.openxma.dsl.pom.model.EventMapping;
import org.openxma.dsl.pom.model.EventType;
import org.openxma.dsl.pom.model.GuiElementEventMapping;
import org.openxma.dsl.pom.model.IGuiElementWithEvent;
import org.openxma.dsl.pom.model.IReferenceableByStatementVariable;
import org.openxma.dsl.pom.model.Image;
import org.openxma.dsl.pom.model.InvokeStatement;
import org.openxma.dsl.pom.model.Label;
import org.openxma.dsl.pom.model.Page;
import org.openxma.dsl.pom.model.ReferencedXMAPage;
import org.openxma.dsl.pom.model.Statement;
import org.openxma.dsl.pom.model.TabPage;
import org.openxma.dsl.pom.model.Table;
import org.openxma.dsl.pom.model.Text;
import org.openxma.dsl.pom.model.VariableValue;
import org.openxma.dsl.pom.model.XmadslPage;
import org.openxma.dsl.pom.naming.PomNameProvider;
import org.openxma.dsl.pom.scoping.PomElementCollector;

import at.spardat.xma.guidesign.DatePicker;
import at.spardat.xma.guidesign.SimpleCombo;
import at.spardat.xma.guidesign.XMACombo;
import at.spardat.xma.guidesign.XMAPage;
import at.spardat.xma.guidesign.XMAPagingControl;
import at.spardat.xma.guidesign.XMATable;
import at.spardat.xma.guidesign.XMAText;
import at.spardat.xma.guidesign.XMAWidget;
import at.spardat.xma.guidesign.flex.XMAFunction;
import at.spardat.xma.guidesign.impl.XMATableImpl;

public class DomPageLogicAttachmentStrategy extends AbstractPomDslAttachmentStrategySupport {

    public static String unEscape(String s) {
        if (s != null) {
            return s.lastIndexOf("^")==-1 ? s : s.replaceAll("\\^", "");
        }
        return null;
    }

    @Override
    public String genAdditionalImportsServer(XMAPage xmaPage) {
        LinkedHashSet<String> imports = new LinkedHashSet<String>();
        EList<Dependant> dependencies = getDependencies(xmaPage);
        if (dependencies != null && dependencies.size() > 0) {
            for (Dependant dependant : dependencies) {
                   imports.add(getFullyQualifiedName(dependant, false));
            }
        }
        new TableCustomizerStrategy().genAdditionalImportsServer(this,xmaPage,imports);
        if (imports.size()>0) {
            StringBuilder stringBuilder = new StringBuilder();
            for (String s : imports) {
                stringBuilder.append("import "+unEscape(s)+";\n");
            }
            return stringBuilder.toString();
        }
        return "";
        }   

    protected boolean canHaveDslWidgetModel(XMAWidget xmaWidget) {
        if (xmaWidget instanceof XMAText) return true;
        if (xmaWidget instanceof SimpleCombo) return true;
        if (xmaWidget instanceof XMACombo) return true;
        if (xmaWidget instanceof DatePicker) return true;
        return false;
    }

    @Override
    public String genAdditionalImportsClient(XMAPage xmaPage) {
        LinkedHashSet<String> imports = new LinkedHashSet<String>();
        List<GuiElementEventMapping> gemList = getCustomGuiElementEventMapping(xmaPage);
        if (gemList.size() >0) {
             imports.add("org.eclipse.swt.graphics.Cursor");
             imports.add("org.eclipse.swt.graphics.Rectangle");
        }
        List<XMAWidget> xmaElements = PomElementCollector.getWidgetsOfXMAPage(xmaPage);
        new TableCustomizerStrategy().genAdditionalImportsClient(this,xmaPage,imports);
        BrowseModeStrategy.getInstance().genAdditionalImportsClient(this,xmaElements,imports);
        TableColumnVisibilityStrategy.getInstance().genAdditionalImportsClient(xmaPage,imports);   
        EffectStrategy.getInstance().genAdditionalImportsClient(xmaPage,imports);          
        InvokeStrategy.genAdditionalImportsClient(xmaPage,imports);
        if (imports.size()>0) {
            StringBuilder stringBuilder = new StringBuilder();
            for (String s : imports) {
                stringBuilder.append("import "+unEscape(s)+";\n");
            }
            return stringBuilder.toString();
        }
     
        
        return "";
    }

    @Override
    public String genAdditionalMemberVariablesClient(XMAPage xmaPage) {
        StringBuilder stringBuilder = new StringBuilder();
        new TableCustomizerStrategy().genAdditionalMemberVariablesClient(this,xmaPage,stringBuilder);
        BrowseModeStrategy.getInstance().genAdditionalMemberVariablesClient(stringBuilder, xmaPage);
        InvokeStrategy.genAdditionalMemberVariablesClient(xmaPage,stringBuilder);
        return stringBuilder.toString();
    }

    @Override
    public String genAdditionalMemberVariablesServer(XMAPage xmaPage) {
        StringBuilder stringBuilder = new StringBuilder();
        EList<Dependant> dependencies = getDependencies(xmaPage);
        if (dependencies != null && dependencies.size() > 0) {
            stringBuilder.append("\n\t/*\n\t * Injected services\n\t */\n");
            for (Dependant dependant : dependencies) {
                stringBuilder.append("\t@org.springframework.beans.factory.annotation.Autowired\n").append("\t").append(getFullyQualifiedName(dependant, false))
                .append(" ").append(Strings.uncapitalize(getName(dependant, false))).append(";\n");
            }
        }
        new TableCustomizerStrategy().genAdditionalMemberVariablesServer(this,xmaPage,stringBuilder);
        return stringBuilder.toString();
    }

    protected EList<Dependant> getDependencies(XMAPage xmaPage) {
        Model model = getReferencingPomModel(xmaPage);
        if (model != null) {
            Iterator<EObject> it = model.eAllContents();
            while (it.hasNext()) {
                EObject obj = it.next();
                if (obj instanceof ReferencedXMAPage && ((ReferencedXMAPage) obj).getXmaPage() != null) {
                    ReferencedXMAPage referencedXMAPage    = (ReferencedXMAPage) obj;
                    if (referencedXMAPage.getXmaPage() != null && referencedXMAPage.getXmaPage().getNamClass().equals(xmaPage.getNamClass())) {
                        return referencedXMAPage.getDependencies();
                    }
                } else if (obj instanceof XmadslPage) {
                    XmadslPage xmadslPage = (XmadslPage)obj;
                    String namClass = Strings.capitalize(xmadslPage.getName());
                    if (namClass.equalsIgnoreCase(xmaPage.getNamClass())) {
                        return xmadslPage.getDependencies();
                    }
                }
            }
        }
        return null;
    }

    public String genAdditionalClientEventCodeClient(XMAPage page, boolean previousEventsExists) {
         List<GuiElementEventMapping> gemList = getCustomGuiElementEventMapping(page);
         if (gemList.size() >0) {
             StringBuilder buf = new StringBuilder();
             for (GuiElementEventMapping gem : gemList) {
                 boolean isDefaultSelection = gem.getEvent().getValue() == EventType.ON_DOUBLE_CLICK_VALUE || gem.getEvent().getValue() == EventType.ON_DEFAULT_SELECTION_VALUE;
                 String defSelString = (isDefaultSelection) ? " && type == SWT.DefaultSelection" : "";
                 if (previousEventsExists) {
                     buf.append(" else ");
                 } else {
                     buf.append("        ");
                 }

                 String widgetName;
                 if (gem.getControl() instanceof CustomizeableField) {
                     widgetName =PomNameProvider.XMA_ID.apply(gem.getControl())+"W";
                 } else {
                     widgetName = PomNameProvider.INSTANCE.getFullyQualifiedName(gem.getControl()).getLastSegment();
                 }
                 buf.append("if (event.widget == "+widgetName+defSelString+") {\n");
                 for (EventFunction ef : gem.getEventFunctions()) {
                     buf.append("            "+ef.getName()+"(");
                     if (ef.getCommand() != null) {
                         buf.append("event");
                     }
                     buf.append(");\n");
                 }
                 buf.append("        }");
             }
             buf.append("\n");
             return (buf.length()>0) ? buf.toString() : "";
         }
         return "";
    }

    public static Page getDslPage(XMAPage xmaPage) {
        List<GuiElementEventMapping> guiElementEventMapping = new ArrayList<GuiElementEventMapping>();
        Model model = getReferencingPomModel(xmaPage);
        if (model != null) {
            Iterator<EObject> it = model.eAllContents();
            while (it.hasNext()) {
                EObject obj = it.next();
                Page dslPage = null;
                String namClass = null;
                if (obj instanceof ReferencedXMAPage && ((ReferencedXMAPage) obj).getXmaPage() != null) {
                    ReferencedXMAPage referencedXMAPage    = (ReferencedXMAPage) obj;
                    dslPage=referencedXMAPage;
                    namClass = Strings.capitalize(referencedXMAPage.getXmaPage().getNamClass());
                } else if (obj instanceof XmadslPage) {
                    XmadslPage xmadslPage = (XmadslPage)obj;
                    dslPage=xmadslPage;
                    namClass = Strings.capitalize(xmadslPage.getName());
                } else if (obj instanceof TabPage) {
                    TabPage tabPage = (TabPage)obj;
                    dslPage=tabPage;
                    namClass = Strings.capitalize(tabPage.getName());                    
                } else {
                    continue;
                }
                if (xmaPage != null) {
                    if (namClass.equalsIgnoreCase(xmaPage.getNamClass())) {
                        return dslPage;
                    }
                }
            }
        }
        return null;
    }

    private List<GuiElementEventMapping> getCustomGuiElementEventMapping(XMAPage page) {
        List<GuiElementEventMapping> guiElementEventMapping = new ArrayList<GuiElementEventMapping>();
        Page dslPage = getDslPage(page);
        if (dslPage != null) {
            if (dslPage.getEvents() != null && dslPage.getEvents().getMapping() != null) {
                List<EventMapping> mapping = dslPage.getEvents().getMapping();
                for (EventMapping em : mapping) {
                    if (isCustomEventMapping(em)) {
                        GuiElementEventMapping gem = (GuiElementEventMapping)em;
                        guiElementEventMapping.add(gem);
                    }
                }
            }
        }
        return guiElementEventMapping;
    }

    private boolean isCustomEventMapping(EventMapping em) {
        if (em instanceof GuiElementEventMapping) {
            GuiElementEventMapping guiEventMapping = (GuiElementEventMapping)em;
            IGuiElementWithEvent guiElementWithEvent = guiEventMapping.getControl();
            if (guiElementWithEvent instanceof Image) {
                return true;
            }
            if (guiElementWithEvent instanceof Label) {
                return true;
            }
            if (guiElementWithEvent instanceof Text) {
                return true;
            }
            if (guiElementWithEvent instanceof CustomizeableField) {
                // TODO support CustomizeableField which results in XMAText controls
                if (BasicTypeHelper.useStaticDataLabel((CustomizeableField)guiElementWithEvent)) {
                    return true;
                } else {
                    return false;
                }
            }
        }
        return false;
    }
    
    public String genCreateStyleCodeClient(XMAPage page) {
        StringBuilder buf = new StringBuilder();        
        DefaultWidgetStyling widgetStyling = new DefaultWidgetStyling();
        widgetStyling.genAdditionalCreateWidgetsCodeClient(this,page,buf);        
        return (buf.length()>0) ? buf.toString() : "";        
    }

    public String genAdditionalCreateWidgetsCodeClient(XMAPage page) {
         StringBuilder buf = new StringBuilder();
         List<GuiElementEventMapping> gemList = getCustomGuiElementEventMapping(page);

         if (gemList.size() >0) {
             buf.append("\n        /* Additional SelectionListeners */\n");
             for (GuiElementEventMapping gem : gemList) {
                 IGuiElementWithEvent control = gem.getControl();
                 String xmaName = PomNameProvider.getXmaName(control);
                 String namePostFix = (control instanceof Label || control instanceof Image) ? "" : "W";
                 buf.append("        addClickSelectionListener("+xmaName+namePostFix+",adapter);\n");
             }
             buf.append("\n        /* Cursor settings */\n");
             buf.append("        final Cursor cursorHand = getComposite().getDisplay().getSystemCursor(SWT.CURSOR_HAND);\n");
             for (GuiElementEventMapping gem : gemList) {
                 IGuiElementWithEvent control = gem.getControl();
                 String xmaName = PomNameProvider.getXmaName(control);
                 String namePostFix = (control instanceof Label || control instanceof Image) ? "" : "W";
                 buf.append("        "+xmaName+namePostFix+".setCursor(cursorHand);\n");
             }
         }


         TableCustomizerStrategy customizerStrategy = new TableCustomizerStrategy();
         customizerStrategy.genCreateWidgetsCodeClient(this,page,buf);
         BrowseModeStrategy.getInstance().genCreateWidgetsCodeClient(buf, page);
         EffectStrategy.getInstance().genAdditionalCreateWidgetsCodeClient(this,page,buf);

//         buf.append("\n        /* Create controls which are displayed when widgets are not editable */\n");
//         buf.append("        createBrowseControls();\n");

         return (buf.length()>0) ? buf.toString() : "";
    }

    public String genAdditionalMethodsClient(XMAPage page) {
        StringBuilder buf = new StringBuilder();
         List<GuiElementEventMapping> gemList = getCustomGuiElementEventMapping(page);
         if (gemList.size() >0) {
            buf.append("    /**\n" +
                    "     * This is a temporary solution generated from the textual presentation model\n" +
                    "     * TODO: Refactor generator and libraries\n" +
                    "     * ATTENTION: A combination of click and doubleClick works not correctly!\n" +
                    "     */\n" +
                    "    protected void addClickSelectionListener(final Control control, final EventAdapter adapter) {\n" +
                    "        control.addMouseListener(new MouseListener() {\n" +
                    "            public SelectionEvent createSelectionEvent(MouseEvent mouseEvent, int type) {\n" +
                    "                 Event e = new Event();\n" +
                    "                 e.widget=mouseEvent.widget;\n" +
                    "                 e.type=type;\n" +
                    "                 e.x = mouseEvent.x;\n" +
                    "                 e.y = mouseEvent.y;\n" +
                    "                 e.button = mouseEvent.button;\n" +
                    "                 SelectionEvent selectionEvent = new SelectionEvent(e);\n" +
                    "                 return selectionEvent;\n" +
                    "            }\n" +
                    "            public void mouseUp(MouseEvent mouseEvent) {\n" +
                    "                 Rectangle r = control.getBounds();\n" +
                    "                 if (mouseEvent.x >= 0 && mouseEvent.x < r.width && mouseEvent.y >= 0 && mouseEvent.y < r.height) {\n" +
                    "                    SelectionEvent selectionEvent = createSelectionEvent(mouseEvent,SWT.Selection);\n" +
                    "                    adapter.widgetSelected(selectionEvent);\n" +
                    "                 }\n" +
                    "            }\n" +
                    "            public void mouseDown(MouseEvent mouseEvent) {\n"+
                    "            }\n" +
                    "            public void mouseDoubleClick(MouseEvent mouseEvent) {\n"+
                    "                SelectionEvent selectionEvent = createSelectionEvent(mouseEvent,SWT.DefaultSelection);\n" +
                    "                adapter.widgetDefaultSelected(selectionEvent);\n" +
                    "            }\n" +
                    "        });\n" +
                    "    }\n");
         }

         TableCustomizerStrategy customizerStrategy = new TableCustomizerStrategy();
         customizerStrategy.genAdditionalMethodsClient(page,buf);
         BrowseModeStrategy.getInstance().genAdditionalMethodsClient(buf, page);
         EffectStrategy.getInstance().genAdditionalMethodsClient(buf, page);         
         TableColumnVisibilityStrategy.getInstance().genAdditionalMethodsClient(buf, page);  
         InvokeStrategy.genAdditionalMethodsClientbuf(buf,page);
         return (buf.length() > 0) ? buf.toString() : "";
    }

    public String genAdditionalMethodsServer(XMAPage page) {

        StringBuilder buffer = new StringBuilder();
        buffer.append(super.genAdditionalMethodsServer(page));

        List<XMATable> tables = PomElementCollector.getElementsOfXMAPage(page, XMATable.class);

        for (XMATable table : tables) {
            if (!(table instanceof XMATableImpl)) {
                throw new IllegalArgumentException("The argument table has to be of type 'XMATableImpl' but the instance '"+table.getNamInstance()+"' is of type:'"+table.getClass()+"'");
            }
            XMATableImpl tableImpl = (XMATableImpl)table;
            if (tableImpl.getBusinessDatas().size()>1) {
                buffer.append("\n");
                TableLogicStrategy tableLogicStrategy = new TableLogicStrategy(tableImpl);
                tableLogicStrategy.genAdditionalFillMethod(buffer);
            }
        }

        TableCustomizerStrategy customizerStrategy = new TableCustomizerStrategy();
        customizerStrategy.genGetQueryObject(this,page,buffer);
        return buffer.toString();
    }
    static String toProperCase(String s) {
        return s.substring(0, 1).toUpperCase() +
                   s.substring(1).toLowerCase();
    }

    public String genAdditionalEnterCodeClient(XMAPage page) {
       String enterCodeString = "";
       TableCustomizerStrategy customizerStrategy = new TableCustomizerStrategy();
       Set<XMATable> tablesWithCustomizer = customizerStrategy.getTablesWithCustomizer(page);
       for (XMATable xmaTable : tablesWithCustomizer) {
           Table table = customizerStrategy.getTableByXMATable(xmaTable);
           customizerStrategy.getTableCustomizerInterface(table);
           XMAPagingControl pagingControl = customizerStrategy.getPagingControl(xmaTable);
           enterCodeString += "\n        ((ITableCustomizer) " + pagingControl.getNamInstance()+".getCustomizer()).init();";
           enterCodeString += "\n        ((ITableCustomizer) " + pagingControl.getNamInstance()+".getCustomizer()).applySettings();";

           String nameXMATable = toProperCase(xmaTable.getNamInstance());
           
           enterCodeString += "\n        new org.openxma.addons.ui.table.customizer.xma.client.ContextMenuHelper(this, "+table.getName() +", (org.openxma.addons.ui.table.customizer.xma.client.TableCustomizerComp) "+pagingControl.getNamInstance()+".getCustomizer()";
       }
       return enterCodeString;
    }    
    
    public String genAdditionalStateChangedClient(XMAPage page) {
        StringBuilder buf = new StringBuilder();
        TableColumnVisibilityStrategy.getInstance().genAdditionalStateChangedClient(buf, page);                           
        return buf.toString();
    }
    
    public String genAdditionalResources(XMAPage page) {
        StringBuilder buf = new StringBuilder();
        TableColumnVisibilityStrategy.getInstance().genAdditionalResources(buf, page);                           
        return buf.toString();
    }        
    
    public String genCommandImplementationClient(XMAPage page, XMAFunction function) {
        String result = super.genCommandImplementationClient(page, function); // ""
        
        final StringBuilder sb = new StringBuilder(result); 
        new FunctionStatementInternalIterator(function, page) {
            @Override
            protected boolean visit(Statement stmt, XMAPage page) {
                if (stmt instanceof InvokeStatement) {
                    InvokeStrategy.genInvokeImplementationClient(page,(InvokeStatement) stmt,sb);
                }
                if (stmt instanceof CallStatement) {
                    CallStatement cs = (CallStatement) stmt;
                    if (cs.getJavaCommand() != null) {
                        String javaCommandStr = cs.getJavaCommand();
                        sb.append("        " + javaCommandStr + "(");
                        sb.append("event");
                        for (VariableValue var : cs.getParameterValues()) {
                            sb.append(",");
                            if (var.getVariable() != null) {
                                IReferenceableByStatementVariable reference = var.getVariable().getReference();
                                sb.append(getParameterName(reference));
                            } else if (var.getStringLiteral() != null && var.getStringLiteral().length() > 0) {
                                sb.append("\"" + var.getStringLiteral() + "\"");
                            } else {
                                sb.append("\"" + var.getIntegerLiteral() + "\"");
                            }
                        }
                        sb.append(");\n");
                    }
                }
                // maybe more to come here...
                return false;
            }
        }.iterate();
        return sb.toString();
    }

    private String getParameterName(IReferenceableByStatementVariable reference) {
        if (reference instanceof Table) {
               return ((Table) reference).getName();
        } else if (reference instanceof Text) {
            return ((Text) reference).getName();            
        }
        throw new RuntimeException("Unsupported parameter type.");
    }

    public Set getAdditionalUsedStateFlags(XMAPage page) {        
        return TableColumnVisibilityStrategy.getInstance().getAdditionalUsedStateFlags(page);         
            
    }
}
