package org.openxma.dsl.generator.impl.substrategy;

import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.openxma.dsl.common.Strings;
import org.openxma.dsl.core.model.DataTypeAndTypeParameter;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.core.model.TrueLiteral;
import org.openxma.dsl.core.model.ValidatorReference;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.DataView;
import org.openxma.dsl.dom.model.Entity;
import org.openxma.dsl.dom.model.FeatureReference;
import org.openxma.dsl.dom.model.ValueObject;
import org.openxma.dsl.generator.helper.BasicType;
import org.openxma.dsl.generator.helper.BasicTypeHelper;
import org.openxma.dsl.generator.helper.EventExtension;
import org.openxma.dsl.generator.helper.WidgetExtension;
import org.openxma.dsl.generator.impl.DomPageLogicAttachmentStrategy;
import org.openxma.dsl.pom.PomPackage;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.StylePropertyTableCustomizer;
import org.openxma.dsl.pom.model.StyleSpecificationProperty;
import org.openxma.dsl.pom.model.Table;
import org.openxma.dsl.pom.model.TableColumn;
import org.openxma.dsl.pom.scoping.PomElementCollector;
import org.openxma.dsl.pom.util.StyleAccess;

import at.spardat.xma.guidesign.XMAPage;
import at.spardat.xma.guidesign.XMAPagingControl;
import at.spardat.xma.guidesign.XMATable;
import at.spardat.xma.guidesign.XMATableColumn;
import at.spardat.xma.guidesign.types.NamedFlag;

public class TableCustomizerStrategy {

    public List<XMATable> getXMATables(XMAPage page) {
        return PomElementCollector.getElementsOfXMAPage(page,XMATable.class);
    }

    public Set<XMATable> getTablesWithCustomizer(XMAPage page) {
        Set<XMATable> pageableTables = getPageableTables(page);
        Set<XMATable> resultSet = new LinkedHashSet<XMATable>();
        for (XMATable xmaTable : pageableTables) {
            Table table = getTableByXMATable(xmaTable);
            String tableCustomizerInterface = null;
            tableCustomizerInterface = getTableCustomizerInterface(table);
            if (tableCustomizerInterface != null) {
                resultSet.add(xmaTable);
            }
        }
        return resultSet;
    }
    public Set<XMATable> getTablesWithExporter(XMAPage page) {
        Set<XMATable> pageableTables = getPageableTables(page);
        Set<XMATable> resultSet = new LinkedHashSet<XMATable>();
        for (XMATable xmaTable : pageableTables) {
            Table table = getTableByXMATable(xmaTable);
            String tableCustomizerInterface = null;
            tableCustomizerInterface = getTableCustomizerInterface(table);
            if (tableCustomizerInterface != null) {
                resultSet.add(xmaTable);
            }
        }
        return resultSet;
    }

    public Set<XMATable> getPageableTables(XMAPage page) {
        List<XMATable> tablesOfPage = getXMATables(page);
        Set<XMATable> resultSet = new LinkedHashSet<XMATable>();
        for (XMATable xmaTable : tablesOfPage) {
            Table table = getTableByXMATable(xmaTable);
            if (table != null) {
                if (table.getPageable() instanceof TrueLiteral) {
                    resultSet.add(xmaTable);
                }
            }
        }
        return resultSet;

    }


    public void genCreateWidgetsCodeClient(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage page, StringBuilder buf) {
        Set<XMATable> tablesWithCustomizer = getTablesWithCustomizer(page);
        for (XMATable xmaTable : tablesWithCustomizer) {

            String tableCustomizerName = xmaTable.getNamInstance() + "_customizer";
            XMAPagingControl pc = getPagingControl(xmaTable);           
            String pcNamInstance = pc.getNamInstance();
            String pcNamInstanceW = pc.getNamInstance() + "W";
            Table table = getTableByXMATable(xmaTable);
            String tableCustomizerInterface = getTableCustomizerInterface(table);

            buf.append("\n        /* setting icon for tableCustomizer on control " + pcNamInstanceW + " */");
            buf.append("\n        " + pcNamInstanceW + ".setCusomizeIcon(org.openxma.addons.ui.table.customizer.common.TableCustomizerUISettingsFactory.createUISettings(this,getComposite()).getCustomizerImage());");

            String xmaTableName = xmaTable.getNamInstance();
            buf.append("\n        /* Table customizer for " + xmaTableName + " */");
            buf.append("\n        " + tableCustomizerInterface + " " + tableCustomizerName + " = (" + tableCustomizerInterface + ") getTypedComponent().getSession().getComponent(" + tableCustomizerInterface + ".COMPONENT_NAME);");
            buf.append("\n        " + tableCustomizerName + ".setTableLayoutManager(" + xmaTable.getNamInstance() + "WLM);");
            buf.append("\n        " + tableCustomizerName + ".setTableWMClient(" + xmaTable.getNamInstance() + ");");
            buf.append("\n        " + tableCustomizerName + ".setFQTableName(\"" + page.getComponent().getNamPackage() + "." + page.getNamClass() + "." + xmaTableName + "\");");
            buf.append("\n        Map<String, ColumnMetaData> " + xmaTableName + "_columnMap = new HashMap<String, ColumnMetaData>();");
            for (Object obj : xmaTable.getColumn()) {
                XMATableColumn xmaColumn = (XMATableColumn) obj;
                TableColumn xmadslColumn = (TableColumn)WidgetExtension.getSourceModelElementProperty(xmaColumn);                
                if (xmadslColumn == null || xmadslColumn.getFeature() == null || xmaColumn.getBdValidator()==null) {
                    continue; // Do not add pure UI columns (without linked
                              // attributes)
                }
                EObject attributeHolder = xmadslColumn.getFeature().getAttributeHolder();
                Attribute attribute = null;
                boolean isValueObject = false;
                if (attributeHolder instanceof FeatureReference) {
                    FeatureReference featureReference = (FeatureReference) attributeHolder;
                    if (featureReference.getSource() instanceof Entity) {
                        attribute = featureReference.getAttribute();
                    }
                } else if (attributeHolder instanceof Attribute) {
                    if (attributeHolder.eContainer() instanceof Entity) {
                        attribute = (Attribute) attributeHolder;
                    } else if (attributeHolder.eContainer() instanceof ValueObject) {
                        attribute = (Attribute) attributeHolder;
                        isValueObject = true;
                    } else if (attributeHolder.eContainer() instanceof DataView) {
                        attribute = (Attribute) attributeHolder;                        
                    }
                    
                }
                List<String> numberTypes = Arrays.asList(new String[]{"Long","Integer","BigDecimal", "Double", "Float"});

                if (attribute != null) {
                    String typeName = null;
                    BasicType basicType = BasicTypeHelper.determineBasicType(attribute);
                    switch (basicType) {
                    case STRING:
                        typeName = "String";
                        break;
                    case I18N_ENUMERATION:
                        typeName= "Enum";
                        break;
                    case BOOLEAN:
                        typeName="Boolean";
                        break;
                    case DATE:
                        typeName="Date";
                        break;
                    case NUMBER:
                        if (numberTypes.contains(attribute.getDataTypeName())) {
                            typeName = attribute.getDataTypeName();
                        } else {
                            ValidatorReference rv = BasicTypeHelper.determineRootValidator(attribute);
                            if (rv.eContainer() instanceof SimpleType) {
                                typeName = ((org.openxma.dsl.core.model.impl.SimpleTypeImpl) rv.eContainer()).getTypeDefinition().getInstanceType();
                                if (typeName.indexOf(".")>0){
                                    typeName = typeName.substring(typeName.lastIndexOf(".")+1);
                                }
                            } else if (attribute.getType().getDataType() instanceof SimpleType){
                                DataTypeAndTypeParameter type = attribute.getType();
                                DataTypeAndTypeParameter rootType = BasicTypeHelper.getRootSimpleType(type); 
                                if (rootType != null) {
                                    typeName = rootType.getDataType().getName();                                    
                                } else {
                                    throw new IllegalArgumentException("type not supported: " + attribute.getDataTypeName());                                    
                                }
                            } else {
                                throw new IllegalArgumentException("type not supported: " + attribute.getDataTypeName());
                            }
                        }                        
                        break;
                    case TIME_STAMP:
                        typeName = "Timestamp";
                        break;
                    case CUSTOM:
                        throw new IllegalArgumentException("type not supported: " + attribute.getDataTypeName());                        
                    default:       
                        typeName = attribute.getDataTypeName();
                        break;
                    }
                    if (isValueObject) {
                        String valuePrefix = ((at.spardat.xma.guidesign.impl.BusinessDataImpl) xmaColumn.getDataAttribute().eContainer()).getNamInstance();
                        buf.append("\n        " + xmaTableName + "_columnMap.put(\"" + valuePrefix + "." + attribute.getName() + "\",new ColumnMetaData(\"" + (valuePrefix + "." + attribute.getName()).toUpperCase()
                                + "\", getGenPageMessages().getString(\"" + xmaColumn.getNamRscKeyLabel() + "\"), " + xmaColumn.getNamInstance().toUpperCase() + ",\"" + typeName + "\"));");
                    } else {
                        buf.append("\n        " + xmaTableName + "_columnMap.put(\"" + attribute.getName() + "\",new ColumnMetaData(\"" + attribute.getName().toUpperCase() + "\", getGenPageMessages().getString(\"" + xmaColumn.getNamRscKeyLabel()
                                + "\"), " + xmaColumn.getNamInstance().toUpperCase() + ",\"" + typeName+ "\"));");
                    }
                }
            }
            buf.append("\n        " + tableCustomizerName + ".setColumnMap(" + xmaTableName + "_columnMap);");
            buf.append("\n        " + tableCustomizerName + ".setAppCode(getAppCode());");
            buf.append("\n        " + tableCustomizerName + ".setPagingWMClient(" + xmaTableName + "_pagingControl);");
            buf.append("\n        " + tableCustomizerName + ".register();");
            buf.append("\n        " + tableCustomizerName + ".createComposite(getComposite().getShell());");
            buf.append("\n        " + xmaTable.getNamInstance() + "_customizerComponentId.set((short)" + tableCustomizerName + ".getId());");
            buf.append("\n        " + pcNamInstance + ".setCustomizer(" + tableCustomizerName + ");");
            buf.append("\n");
            
            //handle CSV Exporter            
            boolean isShowExport = isShowExport(pc);
            if (isShowExport) {
                buf.append("\n        " + pcNamInstanceW + ".setExportIcon(getComponent().getImage(\"at/spardat/xma/mdl/paging/icons/page_excel.png\"));");
                buf.append("\n        " + pcNamInstance + ".setExporter(new Exporter() {");
                buf.append("\n            public void export() {");
                buf.append("\n                ITableCustomizer tc = (ITableCustomizer) " + pcNamInstance + ".getCustomizer();");
                buf.append("\n                CSVWriter csvWriter = null;");
                buf.append("\n                File tmpFile = null;");
                buf.append("\n                try {");
                buf.append("\n                    tmpFile = File.createTempFile(\"" + xmaTableName + "\", \".csv\");");
                buf.append("\n                    csvWriter = new CSVWriter(new FileWriter(tmpFile), ';');");
                buf.append("\n                } catch (IOException e) {");
                buf.append("\n                    throw new AppException(e);");
                buf.append("\n                }");
                buf.append("\n                Map<String, ColumnMetaData> columnMap = tc.getColumnMap();");
                buf.append("\n                CommonTableModel tableModel = tc.getTableModel();");
                buf.append("\n                List<CommonColumnModel> columnModels = tableModel.getColumnModels();");
                buf.append("\n                ");
                buf.append("\n                writeHeader(columnMap, columnModels, csvWriter);");
                buf.append("\n                ");
                buf.append("\n                short id = getNextModelId();");
                buf.append("\n                final int maxResults = " + page.getNamClass() + "Gen.this.get" + Strings.capitalize(xmaTableName)+ "ExportBatchSize();");
                buf.append("\n                try {");
                buf.append("\n                    int count;");
                buf.append("\n                    for (int firstResult = 0; ; firstResult += maxResults) {");
                buf.append("\n                        RemoteCallClient callForPage = newRemoteCall(\"get" + Strings.capitalize(xmaTableName)+ "ExportData\");");
                buf.append("\n                        callForPage.setParameter(0, firstResult);");
                buf.append("\n                        callForPage.setParameter(1, maxResults);");
                buf.append("\n                        RemoteReply " + xmaTableName + "ExportDataReply = callForPage.execute();");
                buf.append("\n                        Collection<?> " + xmaTableName + "ExportData = (Collection<?>) " + xmaTableName + "ExportDataReply.getParameter(0);");
                buf.append("\n                        ");
                buf.append("\n                        count = writeRows(id, " + xmaTableName + "ExportData, columnMap, columnModels, csvWriter);");
                buf.append("\n                        try {");
                buf.append("\n                            csvWriter.flush();");
                buf.append("\n                        } catch (IOException e) {");
                buf.append("\n                            throw new AppException(e);");
                buf.append("\n                        }");
                buf.append("\n                        if (maxResults == 0 || count < maxResults) {");
                buf.append("\n                            break;");
                buf.append("\n                        }");
                buf.append("\n                    }");
                buf.append("\n                } finally {");
                buf.append("\n                    if (csvWriter != null && tmpFile.length() > 0) {");
                buf.append("\n                        try {");
                buf.append("\n                            csvWriter.close();");
                buf.append("\n                            openFile(tmpFile);");
                buf.append("\n                        } catch (IOException e) {");
                buf.append("\n                            throw new AppException(e);");
                buf.append("\n                        }");
                buf.append("\n                    }");
                buf.append("\n                }");
                buf.append("\n            }");
                buf.append("\n");
                buf.append("\n            private int writeRows(short id, Collection<?> exportData, Map<String, ColumnMetaData> columnMap, List<CommonColumnModel> columnModels, CSVWriter csvWriter) {");
                buf.append("\n                int count = 0;");
                buf.append("\n                List<Object> values = new ArrayList<Object>();");
                buf.append("\n                for (Object dto : exportData) {");
                buf.append("\n                    count++;");
                buf.append("\n                    for (CommonColumnModel colModel : columnModels) {");
                buf.append("\n                        if (colModel.isVisible()) {");
                buf.append("\n                            Object attributeValue = getAttributeValue(columnMap, dto, colModel.getNumColumnIndex());");
                buf.append("\n                            IFmt formatter = getFormatter(colModel.getNumColumnIndex());");
                buf.append("\n                            SimpleWMClient c = null;");
                buf.append("\n                            byte type = getTypeByColumn(colModel.getNumColumnIndex());");
                buf.append("\n                            if (attributeValue != null && formatter != null) {");
                buf.append("\n                                c = new SimpleWMClient(id, type, getParent());");
                buf.append("\n                                if (type == Types.T_STRING && !(attributeValue instanceof String)) {");
                buf.append("\n                                    c.set(String.valueOf(attributeValue));");
                buf.append("\n                                } else if (attributeValue instanceof BigDecimal) {");
                buf.append("\n                                    c.set((BigDecimal) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof Integer) {");
                buf.append("\n                                    c.set((Integer) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof Double) {");
                buf.append("\n                                    c.set((Double) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof Float) {");
                buf.append("\n                                    c.set((Float) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof Date) {");
                buf.append("\n                                    c.set((Date) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof Byte) {");
                buf.append("\n                                    c.set((Byte) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof Boolean) {");
                buf.append("\n                                    c.set((Boolean) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof Short) {");
                buf.append("\n                                    c.set((Short) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof String) {");
                buf.append("\n                                    c.set((String) attributeValue);");
                buf.append("\n                                } else if (attributeValue instanceof Long) {");
                buf.append("\n                                    c.set((Long) attributeValue);");
                buf.append("\n                                } else {");
                buf.append("\n                                    c = new SimpleWMClient(id, at.spardat.enterprise.util.Types.T_STRING,");
                buf.append("\n                                            getParent());");
                buf.append("\n                                    c.set(attributeValue.toString());");
                buf.append("\n                                }");
                buf.append("\n                                c.setFmt(formatter);");
                buf.append("\n                                values.add(c.getFormattedString());");
                buf.append("\n                            } else {");
                buf.append("\n                                values.add(attributeValue == null ? \"\" : attributeValue.toString());");
                buf.append("\n                            }");
                buf.append("\n                        }");
                buf.append("\n                    }");
                buf.append("\n                    String[] row = (String[]) values.toArray(new String[values.size()]);");
                buf.append("\n                    values.clear();");
                buf.append("\n                    csvWriter.writeNext(row);");
                buf.append("\n                }");
                buf.append("\n                return count;");
                buf.append("\n            }");
                buf.append("\n");
                buf.append("\n            private void writeHeader(Map<String, ColumnMetaData> columnMap, List<CommonColumnModel> columnModels, CSVWriter csvWriter) {");
                buf.append("\n                List<Object> values = new ArrayList<Object>();");
                buf.append("\n                for (CommonColumnModel colModel : columnModels) {");
                buf.append("\n                    if (colModel.isVisible()) {");
                buf.append("\n                        values.add(colModel.getTxtColumn());");
                buf.append("\n                    }");
                buf.append("\n                }");
                buf.append("\n                String[] header = (String[]) values.toArray(new String[values.size()]);");
                buf.append("\n                csvWriter.writeNext(header);");
                buf.append("\n            }");
                buf.append("\n");
                buf.append("\n            private void openFile(File tmpFile) {");
                buf.append("\n                String[] args = { \"rundll32.exe\", \"url.dll,FileProtocolHandler\", tmpFile.getAbsolutePath() };");
                buf.append("\n                try {");
                buf.append("\n                    Runtime.getRuntime().exec(args);");
                buf.append("\n                } catch (IOException e) {");
                buf.append("\n                    throw new AppException(\"Export failed\", e);");
                buf.append("\n                }");
                buf.append("\n            }");
                buf.append("\n");
                buf.append("\n            private byte getTypeByColumn(int numColumn) {");
                buf.append("\n                byte columnType = " + xmaTableName + ".getColumnType(numColumn);");
                buf.append("\n                if (Types.T_DOM == columnType) {");
                buf.append("\n                    return Types.T_STRING;");
                buf.append("\n                }");
                buf.append("\n                return columnType;");
                buf.append("\n            }");
                buf.append("\n");
                buf.append("\n            private IFmt getFormatter(int numColumn) {");
                buf.append("\n                return ((TableWMClient) " + xmaTableName + ").getColumn(numColumn).getFormatter();");
                buf.append("\n            }");
                buf.append("\n");
                buf.append("\n            private Object getAttributeValue(Map<String, ColumnMetaData> columnMap, Object dto, int modelColumnIndex) {");
                buf.append("\n                String attributeName = getAttributeNameByColumnModelIndex(columnMap, modelColumnIndex);");
                buf.append("\n                String[] attributeHierarchy = attributeName.split(\"\\\\.\");");
                buf.append("\n                String getter = \"\";");
                buf.append("\n                Object attributeValue = dto;");
                buf.append("\n                for (int j = 0; j < attributeHierarchy.length; j++) {");
                buf.append("\n                    getter = \"get\" + attributeHierarchy[j].substring(0, 1).toUpperCase()");
                buf.append("\n                            + attributeHierarchy[j].substring(1, attributeHierarchy[j].length());");
                buf.append("\n                    Method method;");
                buf.append("\n                    try {");
                buf.append("\n                        method = attributeValue.getClass().getMethod(getter);");
                buf.append("\n                        attributeValue = method.invoke(attributeValue);");
                buf.append("\n                    } catch (Exception e) {");
                buf.append("\n                        throw new AppException(e);");
                buf.append("\n                    }");
                buf.append("\n                }");
                buf.append("\n                return attributeValue;");
                buf.append("\n            }");
                buf.append("\n");
                buf.append("\n            private String getAttributeNameByColumnModelIndex(Map<String, ColumnMetaData> columnMap, int modelColumnIndex) {");
                buf.append("\n                for (Entry<String, ColumnMetaData> entry : columnMap.entrySet()) {");
                buf.append("\n                    if ((Integer) entry.getValue().getIndex() == modelColumnIndex) {");
                buf.append("\n                        return entry.getKey();");
                buf.append("\n                    }");
                buf.append("\n                }");
                buf.append("\n                return null;");
                buf.append("\n            }");
                buf.append("\n        });");
                buf.append("\n");
            }
      }            
    }

    private boolean isShowExport(XMAPagingControl pc) {
        boolean isShowExport = false;
        EList<NamedFlag> styles = pc.getCustomStyles().getStyles();            
        for (NamedFlag flag : styles ) {
            if ("SHOW_EXPORT".equals(flag.getName()) && flag.isValue()) {
                isShowExport = true;
            }
        }
        return isShowExport;
    }    

    public XMAPagingControl getPagingControl(XMATable xmaTable) {
        return EventExtension.getPagingControlForXMATable(xmaTable);
    }

    public void genAdditionalImportsClient(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage xmaPage, LinkedHashSet<String> imports) {
        List<XMATable> tablesOfPage = getXMATables(xmaPage);
        for (XMATable xmaTable : tablesOfPage) {
            GuiElement guiElement = WidgetExtension.getGuiElementProperty(xmaTable);
            if (!(guiElement instanceof Table)) continue;
            Table table = (Table)guiElement;
            if (table.getPageable() instanceof TrueLiteral) {
                Map<Integer,StyleSpecificationProperty> styleSpecificationProperties = StyleAccess.getCombinedStyleProperty(table);
                if (styleSpecificationProperties != null) {
                    StylePropertyTableCustomizer customizerProp = getStylePropertyTableCustomizer(styleSpecificationProperties);
                    if (customizerProp != null) {
                        String customizerType = customizerProp.getTableCustomizer().getInstanceType();
                        imports.add(customizerType);
                        imports.add("java.util.Map");
                        imports.add("java.util.HashMap");
                        imports.add("org.openxma.addons.ui.table.customizer.common.ColumnMetaData");
                        XMAPagingControl pc = xmaTable.getPagingControl();
                        if (isShowExport(pc)) {
                            imports.add("at.spardat.xma.rpc.RemoteCallClient");
                            imports.add("at.spardat.xma.rpc.RemoteReply");
                            imports.add("au.com.bytecode.opencsv.CSVWriter");
                            imports.add("java.util.Collection");
                            imports.add("java.util.ArrayList");
                            imports.add("java.util.List");
                            imports.add("java.util.Date");
                            imports.add("java.math.BigDecimal");                            
                            imports.add("java.io.File");
                            imports.add("java.io.FileWriter");
                            imports.add("java.io.IOException");
                            imports.add("java.lang.reflect.Method");
                            imports.add("java.util.Map.Entry");
                            imports.add("at.spardat.enterprise.exc.AppException");
                            imports.add("org.openxma.addons.ui.table.customizer.common.CommonTableModel");
                            imports.add("org.openxma.addons.ui.table.customizer.common.CommonColumnModel");

                        }
                    }
                }
            }
        }
    }

    public void genAdditionalImportsServer(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage xmaPage, LinkedHashSet<String> imports) {
        List<XMATable> tablesOfPage = getXMATables(xmaPage);
        for (XMATable xmaTable : tablesOfPage) {
            GuiElement guiElement = WidgetExtension.getGuiElementProperty(xmaTable);
            if (!(guiElement instanceof Table)) continue;
            Table table = (Table)guiElement;
            if (table.getPageable() instanceof TrueLiteral) {
                Map<Integer,StyleSpecificationProperty> styleSpecificationProperties = StyleAccess.getCombinedStyleProperty(table);
                if (styleSpecificationProperties != null) {
                    StylePropertyTableCustomizer customizerProp = getStylePropertyTableCustomizer(styleSpecificationProperties);
                    if (customizerProp != null) {
                        imports.add("org.openxma.dsl.platform.query.QueryObject");
                        imports.add("at.spardat.xma.session.XMASessionServer");
                        imports.add(getTableCustomizerCompName(customizerProp.getTableCustomizer().getInstanceType(),true));
                        if (isShowExport(xmaTable.getPagingControl())) {
                            imports.add("at.spardat.xma.rpc.RemoteCall");
                            imports.add("at.spardat.xma.rpc.RemoteReply");
                            imports.add("java.util.Collection");
                            imports.add("java.util.Collections");
                        }
                    }
                }
            }
        }
    }

    public void genAdditionalMemberVariablesClient(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage xmaPage, StringBuilder stringBuilder) {
    }

    public void genAdditionalMemberVariablesServer(
            DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy,
            XMAPage xmaPage, StringBuilder stringBuilder) {

    }

    protected String getTableCustomizerCompName(String tableCustomizerInterfaceName, boolean fullQualified) {
        if (tableCustomizerInterfaceName == null) {
            throw new IllegalArgumentException("The value of tableCustomizerInterfaceName must not be null");
        }
        String tableCustomizername;
        int lastDotIndex = tableCustomizerInterfaceName.lastIndexOf('.');
        if (lastDotIndex >= 0) {
            tableCustomizername = tableCustomizerInterfaceName.substring(lastDotIndex+1);
        } else {
            tableCustomizername = tableCustomizerInterfaceName;
        }
        tableCustomizername+="Server";
        if (fullQualified && lastDotIndex >= 0) {
            String clientPackage = tableCustomizerInterfaceName.substring(0,lastDotIndex);
            if (clientPackage.endsWith(".client")) {
                int clientDotIndex = clientPackage.lastIndexOf(".client");
                String serverPackage = clientPackage.substring(0,clientDotIndex)+".server.";
                tableCustomizername = serverPackage+tableCustomizername;
            } else {
                tableCustomizername = clientPackage+"."+tableCustomizername;
            }
        }
        return tableCustomizername;
    }

    
    public void genGetQueryObject(DomPageLogicAttachmentStrategy domPageLogicAttachmentStrategy, XMAPage page, StringBuilder buffer) {

         List<XMATable> tablesOfPage = getXMATables(page);
            for (XMATable xmaTable : tablesOfPage) {
                GuiElement guiElement = WidgetExtension.getGuiElementProperty(xmaTable);
                if (!(guiElement instanceof Table)) continue;
                Table table = (Table)guiElement;
                String tableCustomizername = null;
                if (table.getPageable() instanceof TrueLiteral) {
                    Map<Integer,StyleSpecificationProperty> styleSpecificationProperties = StyleAccess.getCombinedStyleProperty(table);
                    if (styleSpecificationProperties != null) {
                        StylePropertyTableCustomizer customizerProp = getStylePropertyTableCustomizer(styleSpecificationProperties);
                        if (customizerProp != null) {
                            tableCustomizername = getTableCustomizerCompName(customizerProp.getTableCustomizer().getInstanceType(),false);
                        }
                    }
                    if (tableCustomizername == null) {
                        continue;
                    }
                    XMAPagingControl pc = getPagingControl(xmaTable);
                    String xmaTableName = xmaTable.getNamInstance();
                    buffer.append(
                        "\n    /**\n"+
                        "     * Gets the QueryObject for the table <tt>"+xmaTableName+"</tt>\n"+
                        "     */\n"+
                        "     public QueryObject get"+Strings.capitalize(xmaTableName)+"QueryObject() {\n"+
                        "         int offset = "+pc.getNamInstance()+".getOffset();\n"+
                        "         short pageSize = "+pc.getNamInstance()+".getPageSize();\n"+

                        "         short sortingColumn = "+pc.getNamInstance()+".getSortingColumn();\n"+
                        "         boolean isAscending = "+pc.getNamInstance()+".getAscending();\n"+
                        "         XMASessionServer xmaSessionServer = (XMASessionServer)XMASessionServer.getXMASession();\n"+
                        "         "+tableCustomizername+" tableCustomizer = ("+tableCustomizername+")xmaSessionServer.getComponent(Short.valueOf("+xmaTableName+"_customizerComponentId.getEncodedValue()));\n"+
                        "         QueryObject filterQueryForTable = null;\n"+
                        "         if (sortingColumn >-1) {\n"+
                        "             filterQueryForTable = tableCustomizer.getQueryObject(sortingColumn,isAscending);\n"+
                        "         } else {\n"+
                        "             filterQueryForTable = tableCustomizer.getQueryObject();\n"+
                        "         }\n"+
                        "         filterQueryForTable.setFirstResult(offset);\n"+
                        "         filterQueryForTable.setMaxResults(pageSize);\n"+
                        "         return filterQueryForTable;\n"+
                        "     }\n");
                    if (isShowExport(pc)) {
                        buffer.append("\n    /**");
                        buffer.append("\n     * Implement this method to query for export data. Be sure to keep within the boundries of {@link QueryObject#getFirstResult()}"); 
                        buffer.append("\n     * and {@link QueryObject#getMaxResults()} if &gt; 0");
                        buffer.append("\n     */");
                        buffer.append("\n    public abstract Collection<?> queryFor" + Strings.capitalize(xmaTableName) + "ExportData(QueryObject queryObject);");
                        buffer.append("\n");
                        buffer.append("\n    public void get" + Strings.capitalize(xmaTableName)  + "ExportData(RemoteCall call, RemoteReply reply) {");
                        buffer.append("\n        int firstResult = 0;");
                        buffer.append("\n        int maxResults = 0;");
                        buffer.append("\n        if (call.getParameter(0) != null && call.getParameter(0) instanceof Integer");
                        buffer.append("\n                && call.getParameter(1) != null && call.getParameter(1) instanceof Integer) {");
                        buffer.append("\n            firstResult = (Integer) call.getParameter(0);");
                        buffer.append("\n            maxResults = (Integer) call.getParameter(1);");
                        buffer.append("\n        }");
                        buffer.append("\n        reply.setParameter(0, read" + Strings.capitalize(xmaTableName) +  "ExportData(firstResult, maxResults));");
                        buffer.append("\n     }");
                        buffer.append("\n");
                        buffer.append("\n     protected Collection<?> read" + Strings.capitalize(xmaTableName) +"ExportData(int firstResult, int maxResults) {");
                        buffer.append("\n       Collection<?> result = Collections.emptyList();");
                        buffer.append("\n       short sortingColumn = " + xmaTableName +"_pagingControl.getSortingColumn();");
                        buffer.append("\n       boolean isAscending = " + xmaTableName +"_pagingControl.getAscending();");
                        buffer.append("\n       XMASessionServer xmaSessionServer = (XMASessionServer) XMASessionServer.getXMASession();");
                        buffer.append("\n       ITableCustomizerServer tableCustomizer = (ITableCustomizerServer) xmaSessionServer.getComponent(Short.valueOf(" + xmaTableName + "_customizerComponentId.getEncodedValue()));");
                        buffer.append("\n       QueryObject filterQueryForTable = null;");
                        buffer.append("\n       if (sortingColumn > -1) {");
                        buffer.append("\n         filterQueryForTable = tableCustomizer.getQueryObject(sortingColumn, isAscending);");
                        buffer.append("\n       } else {");
                        buffer.append("\n         filterQueryForTable = tableCustomizer.getQueryObject();");
                        buffer.append("\n       }");
                        buffer.append("\n       filterQueryForTable.setFirstResult(firstResult);");
                        buffer.append("\n       filterQueryForTable.setMaxResults(maxResults);");
                        buffer.append("\n       result = queryFor" + Strings.capitalize(xmaTableName) + "ExportData(filterQueryForTable);");
                        buffer.append("\n       return result;");
                        buffer.append("\n     }\n");  
                    }
                }
            }

    }    

    public Table getTableByXMATable(XMATable xmaTable) {
        Table retTable = null;
        GuiElement guiElement = WidgetExtension.getGuiElementProperty(xmaTable);
        if (guiElement instanceof Table){
            retTable = (Table)guiElement;
        }
        return retTable;
    }

    public String getTableCustomizerInterface(Table table) {
        String tableCustomizerInterface = null;
        Map<Integer,StyleSpecificationProperty> styleSpecificationProperties = StyleAccess.getCombinedStyleProperty(table);
        if (styleSpecificationProperties != null) {
            StylePropertyTableCustomizer customizerProp = getStylePropertyTableCustomizer(styleSpecificationProperties);
            if (customizerProp != null) {
                tableCustomizerInterface = customizerProp.getTableCustomizer().getInstanceType();
                int lastDotIndex = tableCustomizerInterface.lastIndexOf('.');
                if (lastDotIndex >= 0) {
                    tableCustomizerInterface = tableCustomizerInterface.substring(lastDotIndex+1);
                }
            }
        }
        return tableCustomizerInterface;
    }

    private StylePropertyTableCustomizer getStylePropertyTableCustomizer(
            Map<Integer, StyleSpecificationProperty> styleSpecificationProperties) {
        return (StylePropertyTableCustomizer)styleSpecificationProperties.get(PomPackage.STYLE_PROPERTY_TABLE_CUSTOMIZER);
    }

    public void genAdditionalMethodsClient(XMAPage page, StringBuilder buf) {
        Set<XMATable> tablesWithCustomizer = getTablesWithCustomizer(page);
        for (XMATable xmaTable : tablesWithCustomizer) {

            String xmaTableName = xmaTable.getNamInstance();
            XMAPagingControl pc = getPagingControl(xmaTable);
            boolean isShowExport = isShowExport(pc);
            if (isShowExport) {
                buf.append("\n");
                buf.append("\n    /** ");
                buf.append("\n     * Defines max. number of lines per call to be fetched from server for CSV export.");
                buf.append("\n     * A value of <code>0</code> retrieves all rows in a singel call (which may lead to");
                buf.append("\n     * out of memory in case of huge result sets)");
                buf.append("\n     */ ");
                buf.append("\n    protected int get" + Strings.capitalize(xmaTableName)+ "ExportBatchSize() {");
                buf.append("\n        return 0;");
                buf.append("\n    }");
                buf.append("\n");
                
            }
        }
    }
}
