package org.openxma.xmadsl.formatting;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.pom.PomPackage;
import org.openxma.dsl.pom.model.ComplexElement;
import org.openxma.dsl.pom.model.Component;
import org.openxma.dsl.pom.model.DataMapping;
import org.openxma.dsl.pom.model.DataObjectVariable;
import org.openxma.dsl.pom.model.Page;
import org.openxma.dsl.pom.model.ReferencedXMAComposite;
import org.openxma.dsl.pom.model.ReferencedXMAPage;

public class ComponentModelFormatting extends ModelFormatting {

	private static ComponentModelFormatting instance;

	public static ModelFormatting getInstance() {
		if (instance == null) {
			instance = new ComponentModelFormatting();
		}
		return instance;
	}

	@Override
	public ArrayList<String> formatModel(EObject obj) {

		if (obj == null || obj instanceof Model == false) {
			throw new IllegalArgumentException("Element to be formatted has to be an instance of Model: " + obj);
		}

		Model componentModel = (Model) obj;

		ArrayList<String> data = new ArrayList<String>();

		formatImports(componentModel, data);
		data.add("");
		formatReferencedXMAComponent(componentModel, data);
		data.add("");
		return data;
	}

	protected void formatReferencedXMAPage(ReferencedXMAPage refPage, ArrayList<String> output) {

		// Begin of page
		output.add("");
		output.add(indent(1) + "Page xma::" + refPage.getXmaPage().getNamClass() + " {");

		// DataObjectVariables
		List<DataObjectVariable> dovList = refPage.getDataobjects();
		output.add("");
		for (int i = 0; i < dovList.size(); i++) {
			DataObjectVariable bd = dovList.get(i);
			String entityName = bd.getName();
			output.add(indent(2) + "Data " + bd.getType().getName() + " " + entityName);
		}

		// Datamapping
		if (refPage.getDataMapping() != null) {
			output.add("");
			output.add(indent(2) + "DataMapping {");
			List<DataMapping> mappingList = refPage.getDataMapping().getMapping();
			for (DataMapping mapping : mappingList) {
				String xmaName = mapping.getControl().getNamInstance();
				String dataObjectVariableName = mapping.getField().getObject().getName();
				String attributName = mapping.getField().getAttribute().getName();
				output.add(indent(3) + "xma::" + xmaName + " <-> " + dataObjectVariableName + "." + attributName);
			}
			output.add(indent(2) + "}");
		}

		// Composites
		if (refPage.getContent() != null) {
			for (ComplexElement complexElement : refPage.getContent().getContainerItems()) {
				if (complexElement instanceof ReferencedXMAComposite) {
					ReferencedXMAComposite refComposite = (ReferencedXMAComposite) complexElement;
					output.add(indent(2) + "Composite xma::" + refComposite.getXmaComposite().getNamInstance() + " {");
					output.add("");
					output.add(indent(2) + "}");
					output.add("");
				}
			}
		}

		// End of page
		output.add(indent(1) + "}");
	}

	/**
	 * adds the collected references to the given string list
	 * 
	 * @param output
	 */
	protected void formatReferencedXMAComponent(Model componentModel, ArrayList<String> output) {

		Component component = (Component) EcoreUtil.getObjectByType(componentModel.getElements(),
				PomPackage.eINSTANCE.getComponent());
		if (componentModel == null) {
			return;
		}
		output.add("Component xma::" + component.getXmaComponent().getNamClass() + " {");
		// if (dataObjectVariable.size()>0) {
		// output.add("");
		// for (DataObjectVariable bd : dataObjectVariable) {
		// String entityName = getEntityNameForBd(bd);
		// output.add(indent(1) + "Data "+entityName+" "+bd.getName()+";");
		// }
		// }
		// output.add("");

		for (Page page : component.getPages()) {
			if (page instanceof ReferencedXMAPage) {
				ReferencedXMAPage refPage = (ReferencedXMAPage) page;
				formatReferencedXMAPage(refPage, output);
			}
		}

		output.add("}");
	}

}
